import pickle

import numpy as np
import pytest
import sympy
from sympy import Symbol as symbol

from pharmpy.model import (
    JointNormalDistribution,
    NormalDistribution,
    RandomVariables,
    VariabilityHierarchy,
    VariabilityLevel,
)


def test_normal_rv():
    dist = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    assert dist.names == ('ETA(1)',)
    assert dist.level == 'IIV'
    dist = dist.replace(name='NEW')
    assert dist.names == ('NEW',)
    with pytest.raises(ValueError):
        NormalDistribution.create('X', 'iiv', 0, -1)


def test_joint_normal_rv():
    dist = JointNormalDistribution.create(['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]])
    assert dist.names == ('ETA(1)', 'ETA(2)')
    assert dist.level == 'IIV'
    dist = dist.replace(names=['NEW', 'ETA(2)'])
    assert dist.names == ('NEW', 'ETA(2)')
    with pytest.raises(ValueError):
        JointNormalDistribution.create(['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[-1, 0.1], [0.1, 2]])


def test_eq_rv():
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    assert dist1 == dist2
    dist3 = NormalDistribution.create('ETA(2)', 'iiv', 0, 1)
    assert dist1 != dist3
    dist4 = NormalDistribution.create('ETA(2)', 'iiv', 0, 0.1)
    assert dist3 != dist4


def test_empty_rvs():
    rvs = RandomVariables.create([])
    assert not rvs
    assert rvs.covariance_matrix == sympy.Matrix()


def test_repr_rv():
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    assert repr(dist1) == 'ETA(1) ~ N(0, 1)'
    dist2 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    assert (
        repr(dist2)
        == """⎡ETA(1)⎤    ⎧⎡0⎤  ⎡ 1   0.1⎤⎫
⎢      ⎥ ~ N⎪⎢ ⎥, ⎢        ⎥⎪
⎣ETA(2)⎦    ⎩⎣0⎦  ⎣0.1   2 ⎦⎭"""
    )


def test_repr_latex_rv():
    dist1 = JointNormalDistribution.create(['x', 'y'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]])
    assert (
        dist1._repr_latex_()
        == '$\\displaystyle \\left[\\begin{matrix}x\\\\y\\end{matrix}\\right]\\sim \\mathcal{N} \\left(\\displaystyle \\left[\\begin{matrix}0\\\\0\\end{matrix}\\right],\\displaystyle \\left[\\begin{matrix}1 & 0.1\\\\0.1 & 2\\end{matrix}\\right]\\right)$'  # noqa E501
    )

    dist2 = NormalDistribution.create('x', 'iiv', 0, 1)
    assert (
        dist2._repr_latex_()
        == '$\\displaystyle x\\sim  \\mathcal{N} \\left(\\displaystyle 0,\\displaystyle 1\\right)$'
    )


def test_parameters_rv():
    dist1 = NormalDistribution.create('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    assert dist1.parameter_names == ('OMEGA(2,2)',)


def test_init():
    rvs = RandomVariables.create([])
    assert len(rvs) == 0
    rvs = RandomVariables.create()
    assert len(rvs) == 0


def test_illegal_inits():
    with pytest.raises(TypeError):
        RandomVariables.create([8, 1])
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    with pytest.raises(ValueError):
        RandomVariables.create([dist1, dist2])
    with pytest.raises(TypeError):
        RandomVariables.create([dist1], eta_levels=23)
    with pytest.raises(TypeError):
        RandomVariables.create([dist1], epsilon_levels=23)


def test_len():
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA(2)', 'iiv', 0, 1)
    rvs = RandomVariables.create([dist1, dist2])
    assert len(rvs) == 2


def test_eq():
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA(2)', 'iiv', 0, 1)
    dist3 = NormalDistribution.create('ETA(3)', 'iiv', 0, 1)
    rvs = RandomVariables.create([dist1, dist2])
    rvs2 = RandomVariables.create([dist1])
    assert rvs != rvs2
    rvs3 = RandomVariables.create([dist1, dist3])
    assert rvs != rvs3
    assert hash(rvs) != hash(rvs2)


def test_add():
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA(2)', 'iiv', 0, 0.1)
    dist3 = NormalDistribution.create('ETA(3)', 'center', 0, 0.1)
    rvs1 = RandomVariables.create([dist1])
    rvs2 = RandomVariables.create([dist2])
    rvs3 = rvs1 + rvs2
    assert len(rvs3) == 2
    assert len(rvs1 + dist2) == 2
    assert len(dist1 + rvs2) == 2
    with pytest.raises(TypeError):
        rvs1 + None
    with pytest.raises(TypeError):
        None + rvs1
    assert len([dist2] + rvs1) == 2

    with pytest.raises(ValueError):
        dist3 + rvs1
    with pytest.raises(ValueError):
        rvs1 + dist3

    lev1 = VariabilityLevel('center', reference=True, group='CENTER')
    levs = VariabilityHierarchy([lev1])
    rvs4 = RandomVariables.create([dist3], eta_levels=levs)
    with pytest.raises(ValueError):
        rvs4 + rvs2


def test_getitem():
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA(2)', 'iiv', 0, 0.1)
    rvs = RandomVariables.create([dist1, dist2])
    assert rvs[0] == dist1
    assert rvs[1] == dist2
    assert rvs['ETA(1)'] == dist1
    assert rvs['ETA(2)'] == dist2
    assert rvs[symbol('ETA(1)')] == dist1
    assert rvs[symbol('ETA(2)')] == dist2
    with pytest.raises(IndexError):
        rvs[23]
    with pytest.raises(KeyError):
        rvs['NOKEYOFTHIS']

    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA(2)', 'iiv', 0, 0.1)
    dist3 = NormalDistribution.create('ETA(3)', 'iiv', 0, 0.1)

    with pytest.raises(KeyError):
        dist1[(0, 1)]

    with pytest.raises(IndexError):
        dist1[1:23]

    with pytest.raises(KeyError):
        dist1['ETA(2)']

    with pytest.raises(IndexError):
        dist1[1]

    with pytest.raises(KeyError):
        dist1[None]

    assert len(dist1[0:1]) == 1
    assert len(dist1['ETA(1)']) == 1
    assert len(dist1[0]) == 1

    rvs = RandomVariables.create([dist1, dist2, dist3])
    selection = rvs[['ETA(1)', 'ETA(2)']]
    assert len(selection) == 2
    selection = rvs[1:]
    assert len(selection) == 2

    dist1 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')],
        ],
    )
    rvs = RandomVariables.create([dist1])
    selection = rvs[['ETA(1)']]
    assert isinstance(selection['ETA(1)'], NormalDistribution)

    with pytest.raises(KeyError):
        rvs[None]

    assert isinstance(dist1[['ETA(1)']], NormalDistribution)

    dist2 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)', 'ETA(3)'],
        'iiv',
        [0, 0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)'), symbol('OMEGA(3,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)'), symbol('OMEGA(2,3)')],
            [symbol('OMEGA(3,1)'), symbol('OMEGA(2,3)'), symbol('OMEGA(3,3)')],
        ],
    )

    assert isinstance(dist2[['ETA(1)', 'ETA(3)']], JointNormalDistribution)
    assert len(dist2[0:2]) == 2
    assert dist2['ETA(1)'].names == ('ETA(1)',)
    assert dist2[0].names == ('ETA(1)',)

    with pytest.raises(KeyError):
        dist2[None]

    with pytest.raises(KeyError):
        dist2[['ETA(1)', 'ETA(4)']]

    with pytest.raises(KeyError):
        dist2[['x', 'y', 'z', 'w']]

    with pytest.raises(KeyError):
        dist2[tuple()]

    with pytest.raises(IndexError):
        dist2[4]

    with pytest.raises(KeyError):
        dist2['ETA(23)']


def test_contains():
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA(2)', 'iiv', 0, 0.1)
    dist3 = NormalDistribution.create('ETA(3)', 'iiv', 0, 0.1)
    rvs = RandomVariables.create([dist1, dist2, dist3])
    assert 'ETA(2)' in rvs
    assert 'ETA(4)' not in rvs


def test_names():
    dist1 = NormalDistribution.create('ETA1', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA2', 'iiv', 0, 0.1)
    rvs = RandomVariables.create([dist1, dist2])
    assert rvs.names == ['ETA1', 'ETA2']


def test_epsilons():
    dist1 = NormalDistribution.create('ETA', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA2', 'iiv', 0, 0.1)
    dist3 = NormalDistribution.create('EPS', 'ruv', 0, 0.1)
    rvs = RandomVariables.create([dist1, dist2, dist3])
    assert rvs.epsilons == RandomVariables.create([dist3])
    assert rvs.epsilons.names == ['EPS']


def test_etas():
    dist1 = NormalDistribution.create('ETA', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA2', 'iiv', 0, 0.1)
    dist3 = NormalDistribution.create('EPS', 'ruv', 0, 0.1)
    rvs = RandomVariables.create([dist1, dist2, dist3])
    assert rvs.etas == RandomVariables.create([dist1, dist2])
    assert rvs.etas.names == ['ETA', 'ETA2']


def test_iiv_iov():
    dist1 = NormalDistribution.create('ETA', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA2', 'iov', 0, 0.1)
    dist3 = NormalDistribution.create('EPS', 'ruv', 0, 0.1)
    rvs = RandomVariables.create([dist1, dist2, dist3])
    assert rvs.iiv == RandomVariables.create([dist1])
    assert rvs.iiv.names == ['ETA']
    assert rvs.iov == RandomVariables.create([dist2])
    assert rvs.iov.names == ['ETA2']


def test_subs():
    dist1 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')],
        ],
    )
    dist2 = NormalDistribution.create('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rvs = RandomVariables.create([dist1, dist2])
    rvs = rvs.subs(
        {
            symbol('ETA(2)'): symbol('w'),
            symbol('OMEGA(1,1)'): symbol('x'),
            symbol('OMEGA(3,3)'): symbol('y'),
        }
    )
    assert rvs['ETA(1)'].variance == sympy.ImmutableMatrix(
        [[symbol('x'), symbol('OMEGA(2,1)')], [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')]]
    )
    assert rvs['ETA(3)'].variance == symbol('y')
    assert rvs.names == ['ETA(1)', 'w', 'ETA(3)']
    dist3 = dist2.subs({'ETA(3)': 'X'})
    assert dist3.names == ('X',)


def test_free_symbols():
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, 1)
    dist2 = NormalDistribution.create('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    assert dist2.free_symbols == {symbol('ETA(2)'), symbol('OMEGA(2,2)')}
    rvs = RandomVariables.create([dist1, dist2])
    assert rvs.free_symbols == {symbol('ETA(1)'), symbol('ETA(2)'), symbol('OMEGA(2,2)')}


def test_parameter_names():
    dist1 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')],
        ],
    )
    assert dist1.parameter_names == ('OMEGA(1,1)', 'OMEGA(2,1)', 'OMEGA(2,2)')
    dist2 = NormalDistribution.create('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    assert dist2.parameter_names == ('OMEGA(3,3)',)
    rvs = RandomVariables.create([dist1, dist2])
    assert rvs.parameter_names == ('OMEGA(1,1)', 'OMEGA(2,1)', 'OMEGA(2,2)', 'OMEGA(3,3)')


def test_subs_rv():
    dist = NormalDistribution.create('ETA(1)', 'iiv', 0, symbol('OMEGA(3,3)'))
    dist = dist.subs({symbol('OMEGA(3,3)'): symbol('VAR')})
    assert dist.variance == symbol('VAR')


def test_repr():
    dist1 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    dist2 = NormalDistribution.create('ETA(3)', 'iiv', 2, 1)
    rvs = RandomVariables.create([dist1, dist2])
    res = """⎡ETA(1)⎤    ⎧⎡0⎤  ⎡ 1   0.1⎤⎫
⎢      ⎥ ~ N⎪⎢ ⎥, ⎢        ⎥⎪
⎣ETA(2)⎦    ⎩⎣0⎦  ⎣0.1   2 ⎦⎭
ETA(3) ~ N(2, 1)"""
    assert repr(rvs) == res
    dist3 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'], 'iiv', [sympy.sqrt(sympy.Rational(2, 5)), 0], [[1, 0.1], [0.1, 2]]
    )
    assert (
        repr(dist3)
        == '''            ⎧⎡√10⎤            ⎫
⎡ETA(1)⎤    ⎪⎢───⎥  ⎡ 1   0.1⎤⎪
⎢      ⎥ ~ N⎪⎢ 5 ⎥, ⎢        ⎥⎪
⎣ETA(2)⎦    ⎪⎢   ⎥  ⎣0.1   2 ⎦⎪
            ⎩⎣ 0 ⎦            ⎭'''
    )


def test_repr_latex():
    dist1 = NormalDistribution.create('z', 'iiv', 0, 1)
    dist2 = JointNormalDistribution.create(['x', 'y'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]])
    rvs = RandomVariables.create([dist1, dist2])
    assert (
        rvs._repr_latex_()
        == '\\begin{align*}\n\\displaystyle z & \\sim  \\mathcal{N} \\left(\\displaystyle 0,\\displaystyle 1\\right) \\\\ \\displaystyle \\left[\\begin{matrix}x\\\\y\\end{matrix}\\right] & \\sim \\mathcal{N} \\left(\\displaystyle \\left[\\begin{matrix}0\\\\0\\end{matrix}\\right],\\displaystyle \\left[\\begin{matrix}1 & 0.1\\\\0.1 & 2\\end{matrix}\\right]\\right)\\end{align*}'  # noqa E501
    )


def test_pickle():
    dist1 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    dist2 = NormalDistribution.create('ETA(3)', 'iiv', 2, 1)
    rvs = RandomVariables.create([dist1, dist2])
    pickled = pickle.dumps(rvs)
    obj = pickle.loads(pickled)
    assert obj == rvs


def test_hash():
    dist1 = NormalDistribution.create('ETA(3)', 'iiv', 2, 1)
    dist2 = NormalDistribution.create('ETA(2)', 'iiv', 2, 0)
    assert hash(dist1) != hash(dist2)
    dist3 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    dist4 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(3)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    assert hash(dist3) != hash(dist4)


def test_dict():
    dist1 = NormalDistribution.create('ETA_3', 'iiv', 2, 1)
    d = dist1.to_dict()
    assert d == {
        'class': 'NormalDistribution',
        'name': 'ETA_3',
        'level': 'IIV',
        'mean': 'Integer(2)',
        'variance': 'Integer(1)',
    }
    dist2 = NormalDistribution.from_dict(d)
    assert dist1 == dist2

    dist1 = NormalDistribution.create('ETA_1', 'iiv', 0, 1)
    dist2 = JointNormalDistribution.create(
        ['ETA_2', 'ETA_3'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA11'), symbol('OMEGA21')],
            [symbol('OMEGA21'), symbol('OMEGA22')],
        ],
    )
    rvs = RandomVariables.create([dist1, dist2])
    d = rvs.to_dict()
    assert d == {
        'dists': (
            {
                'class': 'NormalDistribution',
                'name': 'ETA_1',
                'level': 'IIV',
                'mean': 'Integer(0)',
                'variance': 'Integer(1)',
            },
            {
                'class': 'JointNormalDistribution',
                'names': ('ETA_2', 'ETA_3'),
                'level': 'IIV',
                'mean': 'ImmutableDenseMatrix([[Integer(0)], [Integer(0)]])',
                'variance': "ImmutableDenseMatrix([[Symbol('OMEGA11'), Symbol('OMEGA21')], [Symbol('OMEGA21'), Symbol('OMEGA22')]])",  # noqa: E501
            },
        ),
        'eta_levels': {
            'levels': (
                {'name': 'IIV', 'reference': True, 'group': 'ID'},
                {'name': 'IOV', 'reference': False, 'group': 'OCC'},
            )
        },
        'epsilon_levels': {'levels': ({'name': 'RUV', 'reference': True, 'group': None},)},
    }
    rvs2 = RandomVariables.from_dict(d)
    assert rvs == rvs2


def test_nearest_valid_parameters():
    values = {'x': 1, 'y': 0.1, 'z': 2}
    dist1 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [[symbol('x'), symbol('y')], [symbol('y'), symbol('z')]],
    )
    rvs = RandomVariables.create([dist1])
    new = rvs.nearest_valid_parameters(values)
    assert values == new
    values = {'x': 1, 'y': 1.1, 'z': 1}
    new = rvs.nearest_valid_parameters(values)
    assert new == {'x': 1.0500000000000005, 'y': 1.0500000000000003, 'z': 1.050000000000001}

    dist2 = NormalDistribution.create('ETA(3)', 'iiv', 2, 1)
    rvs = RandomVariables.create([dist2])
    values = {symbol('ETA(3)'): 5}
    new = rvs.nearest_valid_parameters(values)
    assert new == values


def test_validate_parameters():
    a, b, c, d = (symbol('a'), symbol('b'), symbol('c'), symbol('d'))
    dist1 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [[a, b], [c, d]],
    )
    dist2 = NormalDistribution.create('ETA(3)', 'iiv', 0.5, d)
    rvs = RandomVariables.create([dist1, dist2])
    params = {'a': 2, 'b': 0.1, 'c': 1, 'd': 23}
    assert rvs.validate_parameters(params)
    params2 = {'a': 2, 'b': 2, 'c': 23, 'd': 1}
    assert not rvs.validate_parameters(params2)
    with pytest.raises(TypeError):
        rvs.validate_parameters({})


def test_sample():
    dist = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [[symbol('a'), symbol('b')], [symbol('b'), symbol('c')]],
    )

    rv1, rv2 = list(map(symbol, dist.names))
    rvs = RandomVariables.create([dist])

    params = {'a': 1, 'b': 0.1, 'c': 2}
    samples = rvs.sample(rv1 + rv2, parameters=params, samples=2, rng=9532)
    assert list(samples) == pytest.approx([1.7033555824617346, -1.4031809274765599])

    with pytest.raises(ValueError):
        rvs.sample(rv1 + rv2, samples=1, rng=np.random.default_rng(9532))

    samples = rvs.sample(1, samples=2)
    assert list(samples) == [1.0, 1.0]

    with pytest.raises(ValueError):
        rvs.sample(symbol('ETA3'), parameters=params)


def test_variance_parameters():
    dist1 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')],
        ],
    )
    dist2 = NormalDistribution.create('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rvs = RandomVariables.create([dist1, dist2])
    assert rvs.variance_parameters == ['OMEGA(1,1)', 'OMEGA(2,2)', 'OMEGA(3,3)']

    dist1 = NormalDistribution.create('x', 'iiv', 0, symbol('omega'))
    dist2 = NormalDistribution.create('y', 'iiv', 0, symbol('omega'))
    rvs = RandomVariables.create([dist1, dist2])
    assert rvs.variance_parameters == ['omega']

    dist3 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(1,1)')],
        ],
    )
    rvs = RandomVariables.create([dist3])
    assert rvs.variance_parameters == ['OMEGA(1,1)']


def test_get_variance():
    dist1 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')],
        ],
    )
    dist2 = NormalDistribution.create('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rvs = RandomVariables.create([dist1, dist2])
    assert rvs['ETA(1)'].get_variance('ETA(1)') == symbol('OMEGA(1,1)')
    assert rvs['ETA(3)'].get_variance('ETA(3)') == symbol('OMEGA(3,3)')

    with pytest.raises(KeyError):
        dist2.get_variance('ETA(5)')


def test_get_covariance():
    dist1 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')],
        ],
    )
    dist2 = NormalDistribution.create('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rvs = RandomVariables.create([dist1, dist2])
    assert rvs.get_covariance('ETA(1)', 'ETA(2)') == symbol('OMEGA(2,1)')
    assert rvs.get_covariance('ETA(3)', 'ETA(2)') == 0

    assert dist2.get_covariance('ETA(3)', 'ETA(3)') == symbol('OMEGA(3,3)')

    with pytest.raises(KeyError):
        dist2.get_covariance('ETA(1)', 'ETA(1)')


def test_unjoin():
    dist1 = JointNormalDistribution.create(
        ['eta1', 'eta2', 'eta3'], 'iiv', [1, 2, 3], [[9, 2, 3], [4, 8, 6], [1, 2, 9]]
    )
    rvs = RandomVariables.create([dist1])
    new = rvs.unjoin('eta1')
    assert new.nrvs == 3
    assert isinstance(new['eta1'], NormalDistribution)
    assert isinstance(new['eta2'], JointNormalDistribution)


def test_join():
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, symbol('OMEGA(1,1)'))
    dist2 = NormalDistribution.create('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    dist3 = NormalDistribution.create('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    dist4 = JointNormalDistribution.create(
        ['ETA(4)', 'ETA(5)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(4,4)'), symbol('OMEGA(5,4)')],
            [symbol('OMEGA(5,4)'), symbol('OMEGA(5,5)')],
        ],
    )
    dist5 = NormalDistribution.create('EPS(1)', 'ruv', 0, symbol('SIGMA(1,1)'))

    rvs = RandomVariables.create([dist1, dist2])
    joined_rvs, _ = rvs.join(['ETA(1)', 'ETA(2)'])
    assert len(joined_rvs) == 1
    assert joined_rvs[0].variance == sympy.Matrix(
        [[symbol('OMEGA(1,1)'), 0], [0, symbol('OMEGA(2,2)')]]
    )

    joined_rvs, _ = rvs.join(['ETA(1)', 'ETA(2)'], fill=1)
    assert joined_rvs[0].variance == sympy.Matrix(
        [[symbol('OMEGA(1,1)'), 1], [1, symbol('OMEGA(2,2)')]]
    )

    joined_rvs, _ = rvs.join(
        ['ETA(1)', 'ETA(2)'], name_template='IIV_{}_IIV_{}', param_names=['CL', 'V']
    )
    assert joined_rvs[0].variance == sympy.Matrix(
        [
            [symbol('OMEGA(1,1)'), symbol('IIV_CL_IIV_V')],
            [symbol('IIV_CL_IIV_V'), symbol('OMEGA(2,2)')],
        ]
    )

    rvs3 = RandomVariables.create([dist1, dist2, dist3])
    joined_rvs, _ = rvs3.join(['ETA(2)', 'ETA(3)'])
    assert joined_rvs[1].variance == sympy.Matrix(
        [[symbol('OMEGA(2,2)'), 0], [0, symbol('OMEGA(3,3)')]]
    )

    rvs5 = RandomVariables.create([dist1, dist2, dist3, dist4, dist5])
    with pytest.raises(KeyError):
        rvs5.join(['ETA(1)', 'ETA(23)'])


def test_parameters_sdcorr():
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, symbol('OMEGA(1,1)'))
    dist2 = NormalDistribution.create('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    rvs = RandomVariables.create([dist1, dist2])
    params = rvs.parameters_sdcorr({'OMEGA(1,1)': 4})
    assert params == {'OMEGA(1,1)': 2}
    params = rvs.parameters_sdcorr({'OMEGA(1,1)': 4, 'OMEGA(2,2)': 16})
    assert params == {'OMEGA(1,1)': 2, 'OMEGA(2,2)': 4}

    dist2 = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [[symbol('x'), symbol('y')], [symbol('y'), symbol('z')]],
    )
    rvs = RandomVariables.create([dist2])
    params = rvs.parameters_sdcorr({'x': 4, 'y': 0.5, 'z': 16, 'k': 23})
    assert params == {'x': 2.0, 'y': 0.0625, 'z': 4.0, 'k': 23}


def test_variability_level():
    level = VariabilityLevel('IIV', reference=True, group='ID')
    assert level.name == 'IIV'
    assert level.reference is True
    assert level.group == 'ID'

    new = level.replace(group='L1')
    assert new.name == 'IIV'
    assert new.reference is True
    assert new.group == 'L1'
    assert hash(level) != hash(new)

    new = level.replace(group='L1', reference=False)
    assert new.name == 'IIV'
    assert new.reference is False
    assert new.group == 'L1'

    assert repr(level) == "VariabilityLevel(IIV, reference=True, group=ID)"

    d = level.to_dict()
    assert d == {'name': 'IIV', 'reference': True, 'group': 'ID'}
    level2 = VariabilityLevel.from_dict(d)
    assert level == level2


def test_variability_hierarchy():
    lev1 = VariabilityLevel('IIV', reference=True, group='ID')
    levs = VariabilityHierarchy((lev1,))
    assert levs[0].name == 'IIV'
    with pytest.raises(IndexError):
        levs[1].name
    lev2 = VariabilityLevel('CENTER', reference=False, group='CENTER')
    levs2 = VariabilityHierarchy((lev2, lev1))
    assert hash(levs) != hash(levs2)
    assert len(levs2) == 2
    lev3 = VariabilityLevel('PLANET', reference=False, group='PLANET')
    levs3 = VariabilityHierarchy((lev3, lev2, lev1))
    assert len(levs3) == 3

    levs4 = levs + lev2
    assert len(levs4) == 2
    levs5 = lev2 + levs
    assert len(levs5) == 2

    d = levs2.to_dict()
    assert d == {
        'levels': (
            {'name': 'CENTER', 'group': 'CENTER', 'reference': False},
            {'name': 'IIV', 'group': 'ID', 'reference': True},
        )
    }
    levs6 = VariabilityHierarchy.from_dict(d)
    assert levs2 == levs6


def test_covariance_matrix():
    dist1 = NormalDistribution.create('ETA(1)', 'iiv', 0, symbol('OMEGA(1,1)'))
    dist2 = NormalDistribution.create('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    rvs = RandomVariables.create([dist1, dist2])
    assert len(rvs.covariance_matrix) == 4

    dist3 = JointNormalDistribution.create(
        ['ETA(3)', 'ETA(4)'],
        'iiv',
        [0, 0],
        [[symbol('x'), symbol('y')], [symbol('y'), symbol('z')]],
    )
    rvs = RandomVariables.create([dist1, dist2, dist3])
    cov = rvs.covariance_matrix
    assert cov == sympy.Matrix(
        [
            [symbol('OMEGA(1,1)'), 0, 0, 0],
            [0, symbol('OMEGA(2,2)'), 0, 0],
            [0, 0, symbol('x'), symbol('y')],
            [0, 0, symbol('y'), symbol('z')],
        ]
    )

    rvs = RandomVariables.create([dist1, dist3, dist2])
    cov = rvs.covariance_matrix
    assert cov == sympy.Matrix(
        [
            [symbol('OMEGA(1,1)'), 0, 0, 0],
            [0, symbol('x'), symbol('y'), 0],
            [0, symbol('y'), symbol('z'), 0],
            [0, 0, 0, symbol('OMEGA(2,2)')],
        ]
    )


def test_get_rvs_with_same_dist():
    var1 = symbol('OMEGA(1,1)')

    dist1 = NormalDistribution.create('ETA1', 'iov', 0, var1)
    dist2 = NormalDistribution.create('ETA2', 'iov', 0, var1)
    dist3 = NormalDistribution.create('ETA3', 'iov', 0, 1)

    rvs = RandomVariables.create([dist1, dist2, dist3])
    same_dist = rvs.get_rvs_with_same_dist('ETA1')
    assert same_dist == RandomVariables.create([dist1, dist2])

    var2 = symbol('OMEGA(2,2)')
    cov = symbol('OMEGA(1,2)')
    cov_matrix = [[var1, cov], [cov, var2]]

    dist1 = JointNormalDistribution.create(['ETA1', 'ETA2'], 'iov', [0, 0], cov_matrix)
    dist2 = JointNormalDistribution.create(['ETA3', 'ETA4'], 'iov', [0, 0], cov_matrix)
    dist3 = NormalDistribution.create('ETA5', 'iov', 0, var1)

    rvs = RandomVariables.create([dist1, dist2, dist3])
    same_dist = rvs.get_rvs_with_same_dist('ETA1')
    assert same_dist == RandomVariables.create([dist1, dist2])


def test_evalf():
    var1 = symbol('OMEGA11')
    dist1 = NormalDistribution.create('ETA1', 'iiv', 0, var1)
    with pytest.raises(ValueError):
        dist1.evalf({})


def test_levels():
    dist1 = NormalDistribution.create('ETA1', 'iiv', 0, 'omega')
    rvs = RandomVariables.create([dist1])
    assert len(rvs.epsilon_levels) == 1
    assert len(rvs.eta_levels) == 2

    assert rvs.eta_levels.levels == {'IIV': 0, 'IOV': 1}
    assert rvs.epsilon_levels.levels == {'RUV': 0}

    with pytest.raises(ValueError):
        rvs.eta_levels + 23

    with pytest.raises(ValueError):
        23 + rvs.eta_levels

    assert len(rvs.eta_levels[['IIV', 'IOV']]) == 2
    assert len(rvs.eta_levels[rvs.eta_levels]) == 2
    with pytest.raises(ValueError):
        rvs.eta_levels[['IOV']]
    assert rvs.eta_levels['IIV'].group == 'ID'
    assert rvs.eta_levels[rvs.eta_levels['IIV']].group == 'ID'

    lev1 = VariabilityLevel('center', reference=True, group='CENTER')
    levs = VariabilityHierarchy([lev1])
    assert levs != rvs.epsilon_levels
    assert rvs.epsilon_levels != rvs.eta_levels
    assert rvs.epsilon_levels != 23

    levs2 = levs.replace(levels=[lev1])
    assert len(levs2) == 1

    lev2 = VariabilityLevel('center', reference=False, group='CENTER')
    with pytest.raises(ValueError):
        VariabilityHierarchy.create([lev2])
    lev3 = VariabilityLevel('other', reference=True, group='OTHER')
    with pytest.raises(ValueError):
        VariabilityHierarchy.create([lev1, lev3])
    with pytest.raises(ValueError):
        VariabilityHierarchy.create([23])
    with pytest.raises(KeyError):
        rvs.eta_levels[None]

    assert len(VariabilityHierarchy.create(levs)) == 1
    assert len(VariabilityHierarchy.create()) == 0

    assert len(lev2 + rvs.eta_levels) == 3

    x1 = VariabilityLevel.create('other', reference=False, group='OTHER')
    x2 = VariabilityLevel.create('iiv', reference=True, group='ID')
    x3 = VariabilityLevel.create('iov', reference=False, group='OCC')
    h = VariabilityHierarchy.create([x1, x2])
    assert h[x2].group == 'ID'
    with pytest.raises(KeyError):
        h[x3]
    with pytest.raises(KeyError):
        h['NOTHING']


def test_replace_with_sympy_rvs():
    var1 = symbol('OMEGA(1,1)')

    dist = NormalDistribution.create('ETA(1)', 'iiv', 0, var1)
    rvs = RandomVariables.create([dist])

    expr_symbs = symbol('ETA(1)') + symbol('x')
    expr_sympy = rvs.replace_with_sympy_rvs(expr_symbs)

    assert expr_symbs != expr_sympy
    assert not all(isinstance(arg, symbol) for arg in expr_sympy.args)

    var2, cov = symbol('OMEGA(2,2)'), symbol('OMEGA(2,1)')

    dist = JointNormalDistribution.create(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[var1, cov], [cov, var2]]
    )
    rvs = RandomVariables.create([dist])

    expr_symbs = symbol('ETA(1)') + symbol('ETA(2)') + symbol('x')
    expr_sympy = rvs.replace_with_sympy_rvs(expr_symbs)

    assert expr_symbs != expr_sympy
    assert not all(isinstance(arg, symbol) for arg in expr_sympy.args)



from __future__ import annotations

from typing import TYPE_CHECKING, List, Mapping, Optional, Union

from pharmpy.internals.expr.eval import eval_expr
from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.expr.subs import subs
from pharmpy.model import Model

from .expressions import (
    calculate_epsilon_gradient_expression,
    calculate_eta_gradient_expression,
    get_individual_prediction_expression,
    get_population_prediction_expression,
)

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
    import sympy
    from scipy import linalg
else:
    from pharmpy.deps import numpy as np
    from pharmpy.deps import pandas as pd
    from pharmpy.deps import sympy
    from pharmpy.deps.scipy import linalg


ParameterMap = Mapping[Union[str, 'sympy.Symbol'], Union[float, 'sympy.Float']]


class DataFrameMapping(Mapping['sympy.Symbol', 'np.ndarray']):
    def __init__(self, df: pd.DataFrame):
        self._df = df

    def __getitem__(self, symbol: sympy.Symbol):
        return self._df[symbol.name].to_numpy()

    def __len__(self):
        return len(self._df)

    def __iter__(self):
        return map(sympy.Symbol, self._df.columns)


def evaluate_expression(
    model: Model,
    expression: Union[str, sympy.Expr],
    parameter_estimates: Optional[ParameterMap] = None,
):
    """Evaluate expression using model

    Calculate the value of expression for each data record.
    The expression can contain dataset columns, variables in model and
    population parameters. If the model has parameter estimates these
    will be used. Initial estimates will be used for non-estimated parameters.

    Parameters
    ----------
    model : Model
        Pharmpy model
    expression : str or sympy expression
        Expression to evaluate
    parameter_estimates : pd.Series
        Parameter estimates to use instead of initial estimates

    Returns
    -------
    pd.Series
        A series of one evaluated value for each data record

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_expression
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> pe = results.parameter_estimates
    >>> evaluate_expression(model, "TVCL*1000", parameter_estimates=pe)
    0      6.573770
    1      6.573770
    2      6.573770
    3      6.573770
    4      6.573770
             ...
    739    5.165105
    740    5.165105
    741    5.165105
    742    5.165105
    743    5.165105
    Length: 744, dtype: float64

    """
    expression = parse_expr(expression)
    full_expr = model.statements.before_odes.full_expression(expression)
    inits = model.parameters.inits
    mapping = inits if parameter_estimates is None else {**inits, **parameter_estimates}
    expr = subs(full_expr, mapping)

    df = model.dataset

    array = eval_expr(expr, len(df), DataFrameMapping(df))
    return pd.Series(array)


def evaluate_population_prediction(
    model: Model, parameters: Optional[ParameterMap] = None, dataset: Optional[pd.DataFrame] = None
):
    """Evaluate the numeric population prediction

    The prediction is evaluated at the current model parameter values
    or optionally at the given parameter values.
    The evaluation is done for each data record in the model dataset
    or optionally using the dataset argument.

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameters : dict
        Optional dictionary of parameters and values
    dataset : pd.DataFrame
        Optional dataset

    Returns
    -------
    pd.Series
        Population predictions

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_population_prediction
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno_linear")
    >>> results = load_example_modelfit_results("pheno_linear")
    >>> pe = results.parameter_estimates
    >>> evaluate_population_prediction(model, parameters=dict(pe))
    0      17.529739
    1      28.179910
    2       9.688648
    3      17.798916
    4      25.023225
             ...
    150    22.459036
    151    29.223295
    152    20.217288
    153    28.472888
    154    34.226455
    Name: PRED, Length: 155, dtype: float64

    See also
    --------
    evaluate_individual_prediction : Evaluate the individual prediction
    """
    y = get_population_prediction_expression(model)
    mapping = model.parameters.inits if parameters is None else parameters
    expr = subs(y, mapping)

    df = model.dataset if dataset is None else dataset

    pred = eval_expr(expr, len(df), DataFrameMapping(df))
    return pd.Series(pred, name='PRED')


def evaluate_individual_prediction(
    model: Model,
    etas: Optional[pd.DataFrame] = None,
    parameters: Optional[ParameterMap] = None,
    dataset: Optional[pd.DataFrame] = None,
):
    """Evaluate the numeric individual prediction

    The prediction is evaluated at the current model parameter values
    or optionally at the given parameter values.
    The evaluation is done for each data record in the model dataset
    or optionally using the dataset argument.
    The evaluation is done at the current eta values
    or optionally at the given eta values.

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model
    etas : dict
        Optional dictionary of eta values
    parameters : dict
        Optional dictionary of parameters and values
    dataset : pd.DataFrame
        Optional dataset

    Returns
    -------
    pd.Series
        Individual predictions

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_individual_prediction
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno_linear")
    >>> results = load_example_modelfit_results("pheno_linear")
    >>> etas = results.individual_estimates
    >>> evaluate_individual_prediction(model, etas=etas)
    0      17.771084
    1      28.881859
    2      11.441728
    3      21.113050
    4      29.783055
             ...
    150    25.375041
    151    31.833395
    152    22.876707
    153    31.905095
    154    38.099690
    Name: IPRED, Length: 155, dtype: float64

    See also
    --------
    evaluate_population_prediction : Evaluate the population prediction
    """

    y = get_individual_prediction_expression(model)
    mapping = model.parameters.inits if parameters is None else parameters
    y = subs(y, mapping)

    df = model.dataset if dataset is None else dataset

    idcol = model.datainfo.id_column.name

    _etas = (
        pd.DataFrame(
            0,
            index=df[idcol].unique(),
            columns=model.random_variables.etas.names,
        )
        if etas is None
        else etas
    )

    _df = df.join(_etas, on=idcol)

    ipred = eval_expr(y, len(_df), DataFrameMapping(_df))
    return pd.Series(ipred, name='IPRED')


def _replace_parameters(model: Model, y: List[sympy.Expr], parameters: Optional[ParameterMap]):
    mapping = model.parameters.inits if parameters is None else parameters
    return [subs(x, mapping) for x in y]


def evaluate_eta_gradient(
    model: Model,
    etas: Optional[pd.DataFrame] = None,
    parameters: Optional[ParameterMap] = None,
    dataset: Optional[pd.DataFrame] = None,
):
    """Evaluate the numeric eta gradient

    The gradient is evaluated at the current model parameter values
    or optionally at the given parameter values.
    The gradient is done for each data record in the model dataset
    or optionally using the dataset argument.
    The gradient is done at the current eta values
    or optionally at the given eta values.

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model
    etas : dict
        Optional dictionary of eta values
    parameters : dict
        Optional dictionary of parameters and values
    dataset : pd.DataFrame
        Optional dataset

    Returns
    -------
    pd.DataFrame
        Gradient

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_eta_gradient
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno_linear")
    >>> results = load_example_modelfit_results("pheno_linear")
    >>> etas = results.individual_estimates
    >>> evaluate_eta_gradient(model, etas=etas)
         dF/dETA_1  dF/dETA_2
    0     -0.159537 -17.609116
    1     -9.325893 -19.562289
    2     -0.104417 -11.346161
    3     -4.452951 -16.682310
    4    -10.838840 -18.981836
    ..          ...        ...
    150   -5.424423 -19.973013
    151  -14.497185 -17.344797
    152   -0.198714 -22.697161
    153   -7.987731 -23.941806
    154  -15.817067 -22.309945
    <BLANKLINE>
    [155 rows x 2 columns]

    See also
    --------
    evaluate_epsilon_gradient : Evaluate the epsilon gradient
    """

    y = calculate_eta_gradient_expression(model)
    y = _replace_parameters(model, y, parameters)

    df = model.dataset if dataset is None else dataset
    idcol = model.datainfo.id_column.name

    if etas is not None:
        _etas = etas
    elif model.initial_individual_estimates is not None:
        _etas = model.initial_individual_estimates
    else:
        _etas = pd.DataFrame(
            0,
            index=df[idcol].unique(),
            columns=model.random_variables.etas.names,
        )

    derivative_names = [f'dF/d{eta}' for eta in model.random_variables.etas.names]

    _df = df.join(_etas, on=idcol)

    return pd.DataFrame(
        {
            name: eval_expr(expr, len(_df), DataFrameMapping(_df))
            for expr, name in zip(y, derivative_names)
        }
    )


def evaluate_epsilon_gradient(
    model: Model,
    etas: Optional[pd.DataFrame] = None,
    parameters: Optional[ParameterMap] = None,
    dataset: Optional[pd.DataFrame] = None,
):
    """Evaluate the numeric epsilon gradient

    The gradient is evaluated at the current model parameter values
    or optionally at the given parameter values.
    The gradient is done for each data record in the model dataset
    or optionally using the dataset argument.
    The gradient is done at the current eta values
    or optionally at the given eta values.

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model
    etas : dict
        Optional dictionary of eta values
    parameters : dict
        Optional dictionary of parameters and values
    dataset : pd.DataFrame
        Optional dataset

    Returns
    -------
    pd.DataFrame
        Gradient

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_epsilon_gradient
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno_linear")
    >>> results = load_example_modelfit_results("pheno_linear")
    >>> etas = results.individual_estimates
    >>> evaluate_epsilon_gradient(model, etas=etas)
         dY/dEPS_1
    0     17.771084
    1     28.881859
    2     11.441728
    3     21.113050
    4     29.783055
    ..          ...
    150   25.375041
    151   31.833395
    152   22.876707
    153   31.905095
    154   38.099690
    <BLANKLINE>
    [155 rows x 1 columns]

    See also
    --------
    evaluate_eta_gradient : Evaluate the eta gradient
    """

    y = calculate_epsilon_gradient_expression(model)
    y = _replace_parameters(model, y, parameters)
    eps_names = model.random_variables.epsilons.names
    repl = {sympy.Symbol(eps): 0 for eps in eps_names}
    y = [subs(x, repl) for x in y]

    df = model.dataset if dataset is None else dataset

    idcol = model.datainfo.id_column.name

    if etas is not None:
        _etas = etas
    elif model.initial_individual_estimates is not None:
        _etas = model.initial_individual_estimates
    else:
        _etas = pd.DataFrame(
            0,
            index=df[idcol].unique(),
            columns=model.random_variables.etas.names,
        )

    _df = df.join(_etas, on=idcol)
    derivative_names = [f'dY/d{eps}' for eps in eps_names]

    return pd.DataFrame(
        {
            name: eval_expr(expr, len(_df), DataFrameMapping(_df))
            for expr, name in zip(y, derivative_names)
        }
    )


def evaluate_weighted_residuals(
    model: Model,
    parameters: Optional[ParameterMap] = None,
    dataset: Optional[pd.DataFrame] = None,
):
    """Evaluate the weighted residuals

    The residuals is evaluated at the current model parameter values
    or optionally at the given parameter values.
    The residuals is done for each data record in the model dataset
    or optionally using the dataset argument.

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameters : dict
        Optional dictionary of parameters and values
    dataset : pd.DataFrame
        Optional dataset

    Returns
    -------
    pd.Series
        WRES

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_weighted_residuals
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno_linear")
    >>> results = load_example_modelfit_results("pheno_linear")
    >>> parameters = results.parameter_estimates
    >>> evaluate_weighted_residuals(model, parameters=dict(parameters))
    0     -0.313859
    1      0.675721
    2     -1.544240
    3      1.921720
    4      1.517677
            ...
    150    1.223935
    151   -0.053334
    152   -0.007023
    153    0.931252
    154    0.778389
    Name: WRES, Length: 155, dtype: float64
    """

    omega = model.random_variables.etas.covariance_matrix
    sigma = model.random_variables.epsilons.covariance_matrix
    parameters = model.parameters.inits if parameters is None else parameters
    omega = omega.subs(parameters)
    sigma = sigma.subs(parameters)
    omega = np.float64(omega)
    sigma = np.float64(sigma)
    df = model.dataset if dataset is None else dataset
    # FIXME: Could have option to gradients to set all etas 0
    etas = pd.DataFrame(
        0,
        index=df[model.datainfo.id_column.name].unique(),
        columns=model.random_variables.etas.names,
    )
    G = evaluate_eta_gradient(model, etas=etas, parameters=parameters, dataset=dataset)
    H = evaluate_epsilon_gradient(model, etas=etas, parameters=parameters, dataset=dataset)
    F = evaluate_population_prediction(model)
    index = df[model.datainfo.id_column.name]
    G.index = index
    H.index = index
    F.index = index
    WRES = np.float64([])
    for i in df[model.datainfo.id_column.name].unique():
        Gi = np.float64(G.loc[[i]])
        Hi = np.float64(H.loc[[i]])
        Fi = F.loc[i:i]
        DVi = (df['DV'][df[model.datainfo.id_column.name] == i]).astype(np.float64).values
        Ci = Gi @ omega @ Gi.T + np.diag(np.diag(Hi @ sigma @ Hi.T))
        WRESi = linalg.sqrtm(linalg.inv(Ci)) @ (DVi - Fi)
        WRES = np.concatenate((WRES, WRESi))

    return pd.Series(WRES, name='WRES')



import pharmpy.model
from pharmpy.deps import sympy
from pharmpy.internals.code_generator import CodeGenerator

from .name_mangle import name_mangle


def add_theta(model: pharmpy.model.Model, cg: CodeGenerator) -> None:
    """
    Add THETAs to code generator when converting a model

    Parameters
    ----------
    model : pharmpy.model.Model
        Pharmpy model object.
    cg : CodeGenerator
        Code generator to add code upon.
    """
    cg.add("# --- THETAS ---")
    thetas = [p for p in model.parameters if p.symbol not in model.random_variables.free_symbols]
    for theta in thetas:
        if model.estimation_steps[0].method not in ["SAEM", "NLME"]:
            add_ini_parameter(cg, theta, boundary=True)
        else:
            add_ini_parameter(cg, theta)


def add_eta(model: pharmpy.model.Model, cg: CodeGenerator, as_list=False) -> None:
    """
    Add ETAs to code generator when converting a model

    Parameters
    ----------
    model : pharmpy.model.Model
        Pharmpy model object.
    cg : CodeGenerator
        Code generator to add code upon.
    as_list : bool
        Add with separation character ","
    """
    cg.add("")
    cg.add("# --- ETAS ---")
    for dist in model.random_variables.etas:
        omega = dist.variance
        if len(dist.names) == 1:
            init = model.parameters[omega.name].init
            code_line = f'{name_mangle(dist.names[0])} ~ {init}'
            if as_list and dist != model.random_variables.etas[-1]:
                code_line += ","
            cg.add(code_line)
        else:
            cg.add(f'{" + ".join([name_mangle(name) for name in dist.names])} ~ c(')
            inits = []
            for row in range(omega.rows):
                for col in range(row + 1):
                    if col == 0 and row != 0:
                        cg.add(f'{", ".join([str(x) for x in inits])},')
                        inits = []
                        inits.append(f'{model.parameters[omega[row, col].name].init}')
                    else:
                        inits.append(model.parameters[omega[row, col].name].init)
            cg.add(f'{", ".join([str(x) for x in inits])}')
            if as_list and dist != model.random_variables.etas[-1]:
                cg.add("),")
            else:
                cg.add(")")


def add_sigma(model: pharmpy.model.Model, cg: CodeGenerator) -> None:
    """
    Add SIGMAs to code generator when converting a model.

    Parameters
    ----------
    model : pharmpy.model.Model
        Pharmpy model object.
    cg : CodeGenerator
        Code generator to add code upon.
    """
    cg.add("")
    cg.add("# --- EPSILONS ---")
    for dist in model.random_variables.epsilons:
        sigma = dist.variance
        if len(dist.names) == 1:
            sigma_param = model.parameters[sigma]
            if sigma_param.init != 1:
                if model.estimation_steps[0].method not in ["SAEM", "NLME"]:
                    add_ini_parameter(cg, sigma_param, boundary=True)
                else:
                    add_ini_parameter(cg, sigma_param)
            elif not sigma_param.fix:
                if model.estimation_steps[0].method not in ["SAEM", "NLME"]:
                    add_ini_parameter(cg, sigma_param, boundary=True)
                else:
                    add_ini_parameter(cg, sigma_param)
        else:
            for row, col in zip(range(sigma.rows), range(sigma.rows + 1)):
                sigma_param = model.parameters[sigma[row, col]]
                if sigma_param.init != 1:
                    if model.estimation_steps[0].method not in ["SAEM", "NLME"]:
                        add_ini_parameter(cg, sigma_param, boundary=True)
                    else:
                        add_ini_parameter(cg, sigma_param)
                elif not sigma_param.fix:
                    if model.estimation_steps[0].method not in ["SAEM", "NLME"]:
                        add_ini_parameter(cg, sigma_param, boundary=True)
                    else:
                        add_ini_parameter(cg, sigma_param)


def add_ini_parameter(cg: CodeGenerator, parameter: sympy.Symbol, boundary: bool = False) -> None:
    """
    Add a parameter to the ini block in nlmixr2. This is performed for theta
    and sigma parameter values as they are handled in the same manner.

    Parameters
    ----------
    cg : CodeGenerator
        Codegenerator object holding the code to be added to.
    parameter : sympy.Symbol
        The parameter to be added. Either theta or sigma
    boundary : bool, optional
        Decide if the parameter should be added with or without parameter
        boundries. The default is False.
    """
    parameter_name = name_mangle(parameter.name)
    if parameter.fix:
        cg.add(f'{parameter_name} <- fixed({parameter.init})')
    else:
        limit = 1000000.0
        if boundary:
            if parameter.lower > -limit and parameter.upper < limit:
                cg.add(
                    f'{parameter_name} <- c({parameter.lower}, {parameter.init}, {parameter.upper})'
                )
            elif parameter.lower <= -limit and parameter.upper < limit:
                cg.add(f'{parameter_name} <- c(-Inf, {parameter.init}, {parameter.upper})')
            elif parameter.lower > -limit and parameter.upper >= limit:
                cg.add(f'{parameter_name} <- c({parameter.lower}, {parameter.init}, Inf)')
            else:
                cg.add(f'{parameter_name} <- {parameter.init}')
        else:
            cg.add(f'{parameter_name} <- {parameter.init}')



import pathlib

import pharmpy.config as config
from pharmpy.model.external.nlmixr import Model, convert_model

from .run import execute_model, parse_modelfit_results, verification

r"""
.. list-table:: Options for the nlmixr plugin
   :widths: 25 25 50 150
   :header-rows: 1

   * - Option name
     - Default value
     - Type
     - Description
   * - ``rpath``
     - ````
     - str
     - Path to R installation directory
"""


class NLMIXRConfiguration(config.Configuration):
    module = 'pharmpy.plugins.nlmixr'
    rpath = config.ConfigItem(
        pathlib.Path(''),
        'Path to R installation directory',
        cls=pathlib.Path,
    )


conf = NLMIXRConfiguration()

__all__ = ('Model', 'convert_model', 'parse_modelfit_results', 'verification', 'execute_model')



from .data import DatasetError, DatasetWarning
from .datainfo import ColumnInfo, DataInfo
from .distributions.symbolic import Distribution, JointNormalDistribution, NormalDistribution
from .estimation import EstimationStep, EstimationSteps, SimulationStep
from .model import Model, ModelError, ModelfitResultsError, ModelSyntaxError
from .parameters import Parameter, Parameters
from .random_variables import RandomVariables, VariabilityHierarchy, VariabilityLevel
from .statements import (
    Assignment,
    Bolus,
    Compartment,
    CompartmentalSystem,
    CompartmentalSystemBuilder,
    Infusion,
    Statement,
    Statements,
    output,
    to_compartmental_system,
)

__all__ = (
    'Assignment',
    'Bolus',
    'ColumnInfo',
    'Compartment',
    'CompartmentalSystem',
    'CompartmentalSystemBuilder',
    'DataInfo',
    'DatasetError',
    'DatasetWarning',
    'Distribution',
    'EstimationStep',
    'EstimationSteps',
    'Infusion',
    'JointNormalDistribution',
    'Model',
    'ModelError',
    'ModelfitResultsError',
    'ModelSyntaxError',
    'NormalDistribution',
    'output',
    'Parameter',
    'Parameters',
    'RandomVariables',
    'SimulationStep',
    'Statement',
    'Statements',
    'to_compartmental_system',
    'VariabilityHierarchy',
    'VariabilityLevel',
)



"""
:meta private:
"""
from typing import Literal

from pharmpy.deps import sympy
from pharmpy.model import (
    Assignment,
    Compartment,
    CompartmentalSystem,
    CompartmentalSystemBuilder,
    Model,
    Statements,
    output,
)
from pharmpy.modeling import get_central_volume_and_clearance, set_initial_condition

from .error import set_proportional_error_model
from .odes import add_individual_parameter, set_initial_estimates

PD_TYPES = ('linear', 'emax', 'sigmoid', 'step', 'loglin')


def add_effect_compartment(model: Model, expr: Literal[PD_TYPES]):
    r"""Add an effect compartment.

    Implemented PD models are:


    * Linear:

        .. math:: E = B \cdot (1 + \text{slope} \cdot C)

    * Emax:

        .. math:: E = B \cdot \Bigg(1 + \frac {E_{max} \cdot C } { EC_{50} + C}  \Bigg)

    * Step effect:

        .. math:: E = \Biggl \lbrace {B \quad \text{if C} \leq 0 \atop B \cdot (1+ E_{max}) \quad \text{else}}

    * Sigmoidal:

        .. math::  E=\Biggl \lbrace {B \cdot \Bigl(1+\frac{E_{max} \cdot C^n}{EC_{50}^n+C^n}\Bigl) \quad \
                    \text{if C}>0 \atop B \quad \text{else}}

    * Log-linear:

        .. math:: E = \text{slope} \cdot \text{log}(C + C_0)

    :math:`B` is the baseline effect

    Parameters
    ----------
    model : Model
        Pharmpy model
    expr : {'linear', 'emax', 'sigmoid', 'step', 'loglin'}
        Name of the PD effect function.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = add_effect_compartment(model, "linear")
    >>> model.statements.ode_system.find_compartment("EFFECT")
    Compartment(EFFECT, amount=A_EFFECT(t), input=KE0*A_CENTRAL(t)/V)
    """
    vc, cl = get_central_volume_and_clearance(model)

    odes = model.statements.ode_system
    central = odes.central_compartment
    cb = CompartmentalSystemBuilder(odes)

    ke0 = sympy.Symbol("KE0")
    met = sympy.Symbol('MET')
    model = add_individual_parameter(model, met.name)
    ke0_ass = Assignment(ke0, 1 / met)

    effect = Compartment.create("EFFECT", input=ke0 * central.amount / vc)
    cb.add_compartment(effect)
    cb.add_flow(effect, output, ke0)

    model = model.replace(
        statements=Statements(
            model.statements.before_odes
            + ke0_ass
            + CompartmentalSystem(cb)
            + model.statements.after_odes
        )
    )

    model = _add_effect(model, expr, effect.amount)
    return model


def set_direct_effect(model: Model, expr: Literal[PD_TYPES]):
    r"""Add an effect to a model.

    Implemented PD models are:


    * Linear:

        .. math:: E = B \cdot (1 + \text{slope} \cdot C)

    * Emax:

        .. math:: E = B \cdot \Bigg(1 + \frac {E_{max} \cdot C } { EC_{50} + C}  \Bigg)

    * Step effect:

        .. math:: E = \Biggl \lbrace {B \quad \text{if C} \leq 0 \atop B \cdot (1+ E_{max}) \quad \text{else}}

    * Sigmoidal:

        .. math::  E=\Biggl \lbrace {B \cdot \Bigl(1+\frac{E_{max} \cdot C^n}{EC_{50}^n+C^n}\Bigl) \quad \
                    \text{if C}>0 \atop B \quad \text{else}}

    * Log-linear:

        .. math:: E = \text{slope} \cdot \text{log}(C + C_0)

    :math:`B` is the baseline effect

    Parameters
    ----------
    model : Model
        Pharmpy model
    expr : {'linear', 'emax', 'sigmoid', 'step', 'loglin'}
        Name of PD effect function.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_direct_effect(model, "linear")
    >>> model.statements.find_assignment("E")
          ⎛SLOPE⋅A_CENTRAL(t)    ⎞
        B⋅⎜────────────────── + 1⎟
    E =   ⎝        V             ⎠
    """
    vc, cl = get_central_volume_and_clearance(model)
    conc = model.statements.ode_system.central_compartment.amount / vc

    model = _add_effect(model, expr, conc)

    return model


def _add_effect(model: Model, expr: str, conc):
    e0 = sympy.Symbol("B")
    model = add_individual_parameter(model, e0.name)
    if expr in ["emax", "sigmoid", "step"]:
        emax = sympy.Symbol("E_MAX")
        model = add_individual_parameter(model, emax.name)
    if expr in ["emax", "sigmoid"]:
        ec50 = sympy.Symbol("EC_50")
        model = add_individual_parameter(model, ec50.name)

    # Add effect E
    if expr == "linear":
        s = sympy.Symbol("SLOPE")
        model = add_individual_parameter(model, s.name)
        E = Assignment(sympy.Symbol('E'), e0 * (1 + (s * conc)))
    elif expr == "emax":
        E = Assignment(sympy.Symbol("E"), e0 * (1 + (emax * conc / (ec50 + conc))))
    elif expr == "step":
        E = Assignment(sympy.Symbol("E"), sympy.Piecewise((e0, conc <= 0), (e0 * (1 + emax), True)))
    elif expr == "sigmoid":
        n = sympy.Symbol("N")  # Hill coefficient
        model = add_individual_parameter(model, n.name)
        model = set_initial_estimates(model, {"POP_N": 1})
        E = Assignment(
            sympy.Symbol("E"),
            sympy.Piecewise(
                ((e0 * (1 + (emax * conc**n / (ec50**n + conc**n)))), conc > 0), (e0, True)
            ),
        )
    elif expr == "loglin":
        s = sympy.Symbol("SLOPE")
        model = add_individual_parameter(model, s.name)
        E = Assignment(sympy.Symbol("E"), s * sympy.log(conc + sympy.exp(e0 / s)))
    else:
        raise ValueError(f'Unknown model "{expr}".')

    # Add dependent variable Y_2
    y_2 = sympy.Symbol('Y_2')
    y = Assignment(y_2, E.symbol)
    dvs = model.dependent_variables.replace(y_2, 2)
    model = model.replace(statements=model.statements + E + y, dependent_variables=dvs)

    # Add error model
    model = set_proportional_error_model(model, dv=2, zero_protection=False)

    return model


def add_indirect_effect(
    model: Model,
    expr: Literal['linear', 'emax', 'sigmoid', 'step'],
    prod: bool = True,
):
    r"""Add indirect (turnover) effect

    The concentration :math:`C_c` has an impact on the production or degradation rate of the response  R:

    * Production:

        .. math:: \frac {dR}{dt} = k_{in} \cdot (1 + f(C_c)) - k_{out} \cdot R

    * Degradation:

        .. math:: \frac {dR}{dt} = k_{in} - k_{out} \cdot (1 + f(C_c)) \cdot R

    :math:`k_{in}` and :math:`k_{out}` can either be inhibited or stimulated.
    Baseline :math:`B = R(0) = R_0 = k_{in}/k_{out}`.

    Models:

    * Linear:

        .. math:: f(C_c) = \text{slope} \cdot C_c

    * Emax:

        .. math:: f(C_c) = \frac {E_{max} \cdot C_c } { EC_{50} + C_c }

    * Sigmoidal:

        .. math::  f(C_c) = \frac{E_{max} \cdot C_c^n}{EC_{50}^n+C^n}


    Parameters
    ----------
    model : Model
        Pharmpy model
    prod : bool
        Production (True) (default) or degradation (False)
    expr : {'linear', 'emax', 'sigmoid', 'step'}
        Name of PD effect function.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = add_indirect_effect(model, expr='linear', prod=True)

    """
    vc, cl = get_central_volume_and_clearance(model)
    odes = model.statements.ode_system
    central = odes.central_compartment
    conc_c = central.amount / vc

    response = Compartment.create("RESPONSE")
    a_response = response.amount

    kin = sympy.Symbol("K_IN")
    kout = sympy.Symbol("K_OUT")
    met = sympy.Symbol('MET')
    model = add_individual_parameter(model, met.name)
    b = sympy.Symbol("B")  # baseline
    model = add_individual_parameter(model, b.name)

    kout_ass = Assignment(kout, 1 / met)
    kin_ass = Assignment(kin, kout * b)

    if expr == 'linear':
        s = sympy.Symbol("SLOPE")
        model = add_individual_parameter(model, s.name)
        R = sympy.Symbol("SLOPE") * conc_c
    elif expr == 'emax':
        emax = sympy.Symbol("E_MAX")
        model = add_individual_parameter(model, emax.name)
        ec50 = sympy.Symbol("EC_50")
        model = add_individual_parameter(model, ec50.name)
        R = emax * conc_c / (ec50 + conc_c)
    elif expr == 'sigmoid':
        emax = sympy.Symbol("E_MAX")
        ec50 = sympy.Symbol("EC_50")
        n = sympy.Symbol("N")
        model = set_initial_estimates(model, {"POP_N": 1})
        model = add_individual_parameter(model, n.name)
        model = add_individual_parameter(model, ec50.name)
        model = add_individual_parameter(model, emax.name)
        R = emax * conc_c**n / (ec50**n + conc_c**n)
    else:
        raise ValueError(f'Unknown model "{expr}".')

    cb = CompartmentalSystemBuilder(odes)
    if prod:
        response = Compartment.create("RESPONSE", input=kin * (1 + R))
        cb.add_compartment(response)
        cb.add_flow(response, output, kout)
    elif not prod:
        response = Compartment.create("RESPONSE", input=kin)
        cb.add_compartment(response)
        cb.add_flow(response, output, kout * (1 + R))

    model = model.replace(
        statements=Statements(
            model.statements.before_odes
            + kout_ass
            + kin_ass
            + CompartmentalSystem(cb)
            + model.statements.after_odes
        )
    )

    model = set_initial_condition(model, "RESPONSE", b)

    # Add dependent variable Y_2
    y_2 = sympy.Symbol('Y_2')
    y = Assignment(y_2, a_response)
    dvs = model.dependent_variables.replace(y_2, 2)
    model = model.replace(statements=model.statements + y, dependent_variables=dvs)

    # Add error model
    model = set_proportional_error_model(model, dv=2, zero_protection=False)

    return model.update_source()


def set_baseline_effect(model: Model, expr: str = 'const'):
    r"""Create baseline effect model.

    Currently implemented baseline effects are:

    Constant baseline effect (const):

        .. math:: E = B

    Parameters
    ----------
    model : Model
        Pharmpy model
    expr : str
        Name of baseline effect function.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_baseline_effect(model, expr='const')
    >>> model.statements.find_assignment("E")
    E = B
    """
    e0 = sympy.Symbol("B")
    model = add_individual_parameter(model, e0.name)

    E = Assignment(sympy.Symbol('E'), e0)

    # Add dependent variable Y_2
    y_2 = sympy.Symbol('Y_2')
    y = Assignment(y_2, E.symbol)
    dvs = model.dependent_variables.replace(y_2, 2)
    model = model.replace(statements=model.statements + E + y, dependent_variables=dvs)

    # Add error model
    model = set_proportional_error_model(model, dv=2, zero_protection=False)

    return model



import pandas as pd
import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model.external.nonmem.table import CovTable, ExtTable, NONMEMTableFile, PhiTable


def test_nonmem_table(pheno_ext):
    ext_table_file = NONMEMTableFile(pheno_ext)
    ext_table = ext_table_file.table
    assert isinstance(ext_table, ExtTable)
    assert list(ext_table.data_frame.columns) == [
        'ITERATION',
        'THETA(1)',
        'THETA(2)',
        'THETA(3)',
        'OMEGA(1,1)',
        'OMEGA(2,1)',
        'OMEGA(2,2)',
        'SIGMA(1,1)',
        'OBJ',
    ]


def test_ext_table(pheno_ext):
    ext_table_file = NONMEMTableFile(pheno_ext)
    ext_table = ext_table_file.table
    assert isinstance(ext_table, ExtTable)
    assert ext_table.number == 1
    assert ext_table.is_evaluation is False
    assert ext_table.problem == 1
    assert ext_table.subproblem == 0
    assert ext_table.iteration2 == 0
    assert ext_table.method == "First Order Conditional Estimation with Interaction"
    assert ext_table.goal_function == "MINIMUM VALUE OF OBJECTIVE FUNCTION"
    assert ext_table.final_parameter_estimates['THETA(1)'] == 0.00469555
    assert list(ext_table.fixed) == [False, False, False, False, True, False, False]
    assert ext_table.final_ofv == 586.27605628188053
    assert ext_table.initial_ofv == 587.36644134661617


def test_phi_table(pheno_phi):
    phi_table_file = NONMEMTableFile(pheno_phi)
    phi_table = phi_table_file.table
    assert isinstance(phi_table, PhiTable)
    assert phi_table.problem == 1
    assert phi_table.subproblem == 0
    assert phi_table.iteration2 == 0
    assert phi_table.method == "First Order Conditional Estimation with Interaction"
    assert phi_table.goal_function is None


def test_phi_table_phi(testdata):
    phi_table_file = NONMEMTableFile(testdata / "nonmem" / "models" / "pheno_mu_2probs.phi")
    phi_table = phi_table_file.table_no(2)
    assert isinstance(phi_table, PhiTable)
    assert phi_table.problem == 2
    etcs = phi_table.etcs
    id3 = etcs.loc[3]
    assert id3.iloc[0, 0] == 0.078347
    assert id3.iloc[1, 0] == id3.iloc[0, 1] == -0.0154369
    assert id3.iloc[1, 1] == 0.0115108
    assert list(id3.columns) == ['ETA(1)', 'ETA(2)']
    assert list(id3.index) == ['ETA(1)', 'ETA(2)']


def test_cov_table(pheno_cov):
    cov_table_file = NONMEMTableFile(pheno_cov)
    cov_table = cov_table_file.table
    assert isinstance(cov_table, CovTable)
    df = cov_table.data_frame
    paramnames = ['THETA(1)', 'THETA(2)', 'THETA(3)', 'OMEGA(1,1)', 'OMEGA(2,2)', 'SIGMA(1,1)']
    assert list(df.columns) == paramnames
    assert list(df.index) == paramnames
    assert df.values[0][0] == 4.41151e-8
    assert df.values[0][3] == -1.09343e-6


def test_create_phi_table(tmp_path):
    with chdir(tmp_path):
        df = pd.DataFrame({'ETA(1)': [1, 2], 'ETA(2)': [5, 6]}, index=[1, 2])
        df.index.name = 'ID'
        phi = PhiTable(df=df)
        correct = (
            ' SUBJECT_NO   ID           ETA(1)       ETA(2)      \n'
            '            1            1  1.00000E+00  5.00000E+00\n'
            '            2            2  2.00000E+00  6.00000E+00\n'
        )
        assert phi.content == correct

        table_file = NONMEMTableFile(tables=[phi])
        table_file.write('my.phi')
        with open('my.phi', 'r') as fp:
            cont = fp.read()
        correct = 'TABLE NO.     1\n' + correct
        assert cont == correct


def test_errors(testdata):
    ext_file = NONMEMTableFile(testdata / 'nonmem' / 'errors' / 'no_header_error.ext')
    ext_table = ext_file.table_no(2)
    assert isinstance(ext_table, ExtTable)
    with pytest.raises(ValueError):
        ext_table.data_frame


def test_nonmemtablefile_notitle_github_issues_1251(tmp_path):
    filename = 'tab'
    with chdir(tmp_path):
        with open(filename, 'w') as fd:
            fd.write(
                ' ID          TIME        CWRES       CIPREDI     VC\n'
                '  1.1000E+02  0.0000E+00  0.0000E+00  0.0000E+00  9.5921E+01\n'
                '  1.1000E+02  0.0000E+00  0.0000E+00  0.0000E+00  9.5921E+01\n'
            )

        table_file = NONMEMTableFile(filename, notitle=True)
        table = table_file[0]
        df = table.data_frame

        assert tuple(df.columns) == ('ID', 'TIME', 'CWRES', 'CIPREDI', 'VC')
        assert len(df) == 2



from __future__ import annotations

from hashlib import sha256
from typing import TYPE_CHECKING, Iterator, Union

if TYPE_CHECKING:
    import pandas as pd
else:
    from pharmpy.deps import pandas as pd


def _pd_hash_values(obj: Union[pd.Index, pd.Series, pd.DataFrame]) -> pd.Series:
    # NOTE: We explicit all arguments for future-proofing
    return pd.util.hash_pandas_object(  # pyright: ignore [reportGeneralTypeIssues]
        obj, index=False, encoding='utf8', hash_key='0123456789123456', categorize=True
    )


def _df_hash_values(df: pd.DataFrame) -> Iterator[pd.Series]:
    yield _pd_hash_values(df.columns)
    yield _pd_hash_values(df.index)
    yield _pd_hash_values(df.dtypes)
    yield _pd_hash_values(df)


def hash_df_runtime(df: pd.DataFrame) -> int:
    return hash(tuple(map(lambda series: tuple(series.values), _df_hash_values(df))))


def hash_df_fs(df: pd.DataFrame) -> str:
    h = sha256()
    for series in _df_hash_values(df):
        h.update(series.to_numpy())  # pyright: ignore [reportGeneralTypeIssues]
    return h.hexdigest()



from collections import defaultdict
from itertools import chain, product
from typing import Callable, Dict, Iterable, Tuple

from pharmpy.model import Model

from .feature.absorption import features as absorption_features
from .feature.covariate import features as covariate_features
from .feature.direct_effect import features as direct_effect_features
from .feature.effect_comp import features as effect_comp_features
from .feature.elimination import features as elimination_features
from .feature.feature import Feature, FeatureFn, FeatureKey
from .feature.indirect_effect import features as indirect_effect_features
from .feature.lagtime import features as lagtime_features
from .feature.metabolite import features as metabolite_features
from .feature.peripherals import features as peripherals_features
from .feature.transits import features as transits_features
from .statement.statement import Statement

FeatureGenerator = Callable[[Model, Iterable[Statement]], Iterable[Feature]]

modelsearch_features = (
    absorption_features,
    elimination_features,
    transits_features,
    peripherals_features,
    lagtime_features,
)
structsearch_pd_features = (direct_effect_features, effect_comp_features, indirect_effect_features)
structsearch_metabolite_features = (metabolite_features, peripherals_features)


def all_funcs(model: Model, statements: Iterable[Statement]):
    return funcs(
        model,
        statements,
        (
            *modelsearch_features,
            covariate_features,
            *structsearch_pd_features,
            *structsearch_metabolite_features,
        ),
    )


def funcs(
    model: Model, statements: Iterable[Statement], generators: Iterable[FeatureGenerator]
) -> Dict[FeatureKey, FeatureFn]:
    statements_list = list(statements)  # TODO: Only read statements once

    features = chain.from_iterable(
        map(
            lambda features: features(model, statements_list),
            generators,
        )
    )

    return dict(features)


def _group_incompatible_features(funcs):
    grouped = defaultdict(list)
    for key in funcs.keys():
        grouped[key[0]].append(key)
    return grouped.values()


def all_combinations(fns: Dict[FeatureKey, FeatureFn]) -> Iterable[Tuple[FeatureKey]]:
    grouped = _group_incompatible_features(fns)
    feats = ((None, *group) for group in grouped)
    for t in product(*feats):
        a = tuple(elt for elt in t if elt is not None)
        if a:
            yield a


def key_to_str(key: FeatureKey) -> str:
    name, *args = key
    return f'{name}({", ".join(map(str, args))})'


def get_funcs_same_type(funcs, feat):
    return [value for key, value in funcs.items() if key[0] == feat[0]]



from pathlib import Path

from .option_record import OptionRecord


class EtasRecord(OptionRecord):
    @property
    def path(self):
        file_option = self.option_pairs['FILE']
        assert file_option is not None
        return Path(file_option)

    def set_path(self, value):
        newrec = self.set_option('FILE', str(value))
        return newrec



from dataclasses import dataclass
from typing import Literal, Tuple, Union

from lark.visitors import Interpreter

from .feature import ModelFeature, feature
from .symbols import Name, Wildcard


@dataclass(frozen=True)
class IndirectEffect(ModelFeature):
    modes: Union[Tuple[Name[Literal['LINEAR', 'EMAX', 'SIGMOID']], ...], Wildcard]
    production: Union[Tuple[Name[Literal['DEGRADATION', 'PRODUCTION']], ...], Wildcard]


class IndirectEffectInterpreter(Interpreter):
    def interpret(self, tree):
        children = self.visit_children(tree)
        assert len(children) == 2
        return feature(IndirectEffect, children)

    def pdtype_modes(self, tree):
        children = self.visit_children(tree)
        return list(Name(child.value.upper()) for child in children)

    def production_modes(self, tree):
        children = self.visit_children(tree)
        return list(Name(child.value.upper()) for child in children)

    def pdtype_wildcard(self, tree):
        return Wildcard()

    def production_wildcard(self, tree):
        return Wildcard()



from __future__ import annotations

from operator import is_
from typing import TYPE_CHECKING, Callable, List, Type

if TYPE_CHECKING:
    import sympy
else:
    from pharmpy.deps import sympy


def replace_root_children(expr: sympy.Expr, args: List[sympy.Expr]):
    # NOTE: This creates a new tree by replacing the children of the root node.
    # If the children have not changed it returns the original tree which
    # allows certain downstream optimizations.
    return expr if all(map(is_, expr.args, args)) else expr.func(*args)


def prune(predicate: Callable[[sympy.Basic], bool], expr: sympy.Expr):
    """Create a new expression by removing subexpressions nodes from an input
    expression.

    Parameters
    ----------
    predicate : (sympy.Basic) -> bool
        A function that takes a subexpression as input and returns a bool
        indicating whether this subexpression should be pruned.
    expr : sympy.Expr
        The input expression.

    Returns
    -------
    sympy.Expr
        The pruned expression.

    Examples
    --------
    >>> import sympy
    >>> from pharmpy.internals.expr.tree import prune
    >>> prune(lambda expr: expr.func is sympy.exp, sympy.exp(2))
    0
    >>> prune(lambda expr: expr.func is sympy.exp, 3*sympy.exp(2))
    3

    See also
    --------
    subs

    """
    if predicate(expr):
        return _neutral(sympy.Add)

    stack = [expr]
    output = [[], []]

    while stack:
        e = stack[-1]
        old_args = e.args

        new_args = output[-1]

        n = len(old_args)
        i = len(new_args)

        if i == n:
            stack.pop()
            output.pop()
            output[-1].append(replace_root_children(e, new_args))
        else:
            old_arg = old_args[i]
            if predicate(old_arg):
                # NOTE: Replace node without recursing
                new_args.append(_neutral(e.func))
            else:
                # NOTE: Push the next argument on the stack
                stack.append(old_arg)  # pyright: ignore [reportGeneralTypeIssues]
                output.append([])

    return output[0][0]


def _neutral(fn: Type[sympy.Basic]) -> sympy.Integer:
    if fn is sympy.Add:
        return sympy.Integer(0)
    if fn is sympy.Mul:
        return sympy.Integer(1)
    if fn is sympy.Pow:
        return sympy.Integer(1)

    raise ValueError(f'{type(fn)}: {repr(fn)}')



from dataclasses import fields
from typing import Iterable, Tuple, Union

from .statement.feature.covariate import Ref
from .statement.feature.symbols import Name, Option, Wildcard
from .statement.statement import Statement

StringifiableAtom = Union[int, str, Name, Wildcard, Ref]
Stringifiable = Union[StringifiableAtom, Tuple[StringifiableAtom, ...]]


def stringify(statements: Iterable[Statement]) -> str:
    return ';'.join(map(_stringify_statement, statements))


def _stringify_statement(statement: Statement) -> str:
    return (
        f'{_stringify_statement_type(statement)}'
        f'{_stringify_optional(statement)}'
        f'({",".join(map(_stringify_attribute, _filter_attributes(statement)))})'
    )


def _stringify_statement_type(statement: Statement) -> str:
    return statement.__class__.__name__.upper()


def _filter_attributes(statement: Statement) -> Iterable[Stringifiable]:
    it = iter(fields(statement))
    for field in it:
        value = getattr(statement, field.name)
        if not isinstance(value, Option):
            if value == field.default:
                break
            yield value

    for field in it:
        # FIXME: Better handling
        value = getattr(statement, field.name)
        if not isinstance(value, Option):
            assert value == field.default


def _stringify_attribute(attribute: Stringifiable) -> str:
    if isinstance(attribute, Ref):
        return f'@{attribute.name}'
    elif isinstance(attribute, Name):
        return attribute.name
    elif isinstance(attribute, Wildcard):
        return '*'
    elif isinstance(attribute, str):
        return attribute
    elif isinstance(attribute, int):
        return str(attribute)
    elif isinstance(attribute, tuple):
        if len(attribute) == 1:
            return _stringify_attribute(attribute[0])
        elif len(attribute) >= 2:
            if type(i := attribute[0]) == type(j := attribute[-1]) == int:  # noqa E721
                assert isinstance(i, int)
                assert isinstance(j, int)
                if tuple(range(i, j + 1)) == attribute:
                    return f'{i}..{j}'

        return f'[{",".join(map(_stringify_attribute, attribute))}]'
    else:
        raise TypeError(type(attribute))


def _stringify_optional(statement: Statement) -> str:
    it = iter(fields(statement))
    for field in it:
        value = getattr(statement, field.name)
        if isinstance(value, Option) and value.option is True:
            return '?'
    return ''



import pandas as pd
import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import NormalDistribution
from pharmpy.modeling import set_seq_zo_fo_absorption
from pharmpy.tools import fit, retrieve_models, run_iivsearch

# FIXME: Tests including modelfit_results are commented out, uncomment once function retrieve_models
#  return model entries or we have a separate function for this


def test_no_of_etas_keep(tmp_path, model_count, start_modelres):
    with chdir(tmp_path):
        res_keep1 = run_iivsearch(
            'brute_force_no_of_etas',
            results=start_modelres[1],
            model=start_modelres[0],
            keep=["CL"],
        )
        no_of_models = 8
        assert len(res_keep1.summary_models) == no_of_models // 2
        assert res_keep1.summary_individuals.iloc[-1]['description'] == '[CL]'


def test_block_structure(tmp_path, model_count, start_modelres):
    with chdir(tmp_path):
        res = run_iivsearch(
            'brute_force_block_structure', results=start_modelres[1], model=start_modelres[0]
        )

        no_of_candidate_models = 4
        assert len(res.summary_tool) == no_of_candidate_models + 1
        assert len(res.summary_models) == no_of_candidate_models + 1

        res_models = [model for model in retrieve_models(res) if model.name != 'input_model']
        assert len(res_models) == no_of_candidate_models

        # assert all(
        #     model.modelfit_results and not np.isnan(model.modelfit_results.ofv)
        #     for model in res.models
        # )
        start_model = start_modelres[0]
        assert all(model.random_variables != start_model.random_variables for model in res_models)

        assert res.summary_tool.loc[1, 'mox2']['description'] == '[CL]+[VC]+[MAT]'
        assert isinstance(start_model.random_variables['ETA_1'], NormalDistribution)

        assert res.summary_tool.loc[1, 'iivsearch_run1']['description'] == '[CL,VC,MAT]'
        assert len(res_models[0].random_variables['ETA_1'].names) == 3

        summary_tool_sorted_by_dbic = res.summary_tool.sort_values(by=['dbic'], ascending=False)
        summary_tool_sorted_by_bic = res.summary_tool.sort_values(by=['bic'])
        summary_tool_sorted_by_rank = res.summary_tool.sort_values(by=['rank'])
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_rank)
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_bic)

        rundir = tmp_path / 'iivsearch_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_candidate_models
        assert (rundir / 'metadata.json').exists()


def test_no_of_etas(tmp_path, model_count, start_modelres):
    with chdir(tmp_path):
        res = run_iivsearch(
            'brute_force_no_of_etas', results=start_modelres[1], model=start_modelres[0]
        )

        no_of_candidate_models = 7
        assert len(res.summary_tool) == no_of_candidate_models + 1
        assert len(res.summary_models) == no_of_candidate_models + 1

        res_models = [model for model in retrieve_models(res) if model.name != 'input_model']
        assert len(res_models) == no_of_candidate_models

        assert res.summary_tool.loc[1, 'mox2']['description'] == '[CL]+[VC]+[MAT]'
        assert start_modelres[0].random_variables.iiv.names == ['ETA_1', 'ETA_2', 'ETA_3']

        assert res.summary_tool.iloc[-1]['description'] == '[]'
        assert res_models[0].random_variables.iiv.names == ['ETA_2', 'ETA_3']

        summary_tool_sorted_by_dbic = res.summary_tool.sort_values(by=['dbic'], ascending=False)
        summary_tool_sorted_by_bic = res.summary_tool.sort_values(by=['bic'])
        summary_tool_sorted_by_rank = res.summary_tool.sort_values(by=['rank'])
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_rank)
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_bic)

        rundir = tmp_path / 'iivsearch_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_candidate_models
        assert (rundir / 'metadata.json').exists()


def test_brute_force(tmp_path, model_count, start_modelres):
    with chdir(tmp_path):
        res = run_iivsearch('brute_force', results=start_modelres[1], model=start_modelres[0])

        no_of_candidate_models = 8
        assert len(res.summary_tool) == no_of_candidate_models + 2
        assert len(res.summary_models) == no_of_candidate_models + 1

        res_models = [model for model in retrieve_models(res) if model.name != 'input_model']
        assert len(res_models) == no_of_candidate_models

        assert 'iivsearch_run3' in res.summary_errors.index.get_level_values('model')

        # assert all(
        #     model.modelfit_results and not np.isnan(model.modelfit_results.ofv)
        #     for model in res.models
        # )
        assert all(
            model.random_variables != start_modelres[0].random_variables for model in res_models
        )

        summary_tool_sorted_by_dbic_step1 = res.summary_tool.loc[1].sort_values(
            by=['dbic'], ascending=False
        )
        summary_tool_sorted_by_bic_step1 = res.summary_tool.loc[1].sort_values(by=['bic'])
        summary_tool_sorted_by_rank_step1 = res.summary_tool.loc[1].sort_values(by=['rank'])
        pd.testing.assert_frame_equal(
            summary_tool_sorted_by_dbic_step1, summary_tool_sorted_by_rank_step1
        )
        pd.testing.assert_frame_equal(
            summary_tool_sorted_by_dbic_step1, summary_tool_sorted_by_bic_step1
        )

        summary_tool_sorted_by_dbic_step2 = res.summary_tool.loc[2].sort_values(
            by=['dbic'], ascending=False
        )
        summary_tool_sorted_by_bic_step2 = res.summary_tool.loc[2].sort_values(by=['bic'])
        summary_tool_sorted_by_rank_step2 = res.summary_tool.loc[2].sort_values(by=['rank'])
        pd.testing.assert_frame_equal(
            summary_tool_sorted_by_dbic_step2, summary_tool_sorted_by_rank_step2
        )
        pd.testing.assert_frame_equal(
            summary_tool_sorted_by_dbic_step2, summary_tool_sorted_by_bic_step2
        )

        rundir = tmp_path / 'iivsearch_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_candidate_models
        assert (rundir / 'metadata.json').exists()


@pytest.mark.filterwarnings("ignore::UserWarning")
@pytest.mark.parametrize(
    'iiv_strategy',
    ['add_diagonal', 'fullblock'],
)
def test_no_of_etas_iiv_strategies(tmp_path, model_count, start_modelres, iiv_strategy):
    with chdir(tmp_path):
        start_model = start_modelres[0].replace(name='moxo2_copy')
        start_model = set_seq_zo_fo_absorption(start_model)
        start_res = fit(start_model)

        res = run_iivsearch(
            'brute_force_no_of_etas',
            iiv_strategy=iiv_strategy,
            results=start_res,
            model=start_model,
        )

        no_of_candidate_models = 15
        assert len(res.summary_tool) == no_of_candidate_models + 1
        assert len(res.summary_models) == no_of_candidate_models + 2

        res_models = [model for model in retrieve_models(res) if model.name != 'input_model']
        assert len(res_models) == no_of_candidate_models + 1

        if iiv_strategy == 'fullblock':
            base_model = [model for model in res_models if model.name == 'base_model'].pop()
            base_rvs = base_model.random_variables.iiv
            assert len(base_rvs['ETA_1']) == base_rvs.nrvs

        summary_tool_sorted_by_dbic = res.summary_tool.sort_values(by=['dbic'], ascending=False)
        summary_tool_sorted_by_bic = res.summary_tool.sort_values(by=['bic'])
        summary_tool_sorted_by_rank = res.summary_tool.sort_values(by=['rank'])
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_rank)
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_bic)

        rundir = tmp_path / 'iivsearch_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_candidate_models + 1
        assert (rundir / 'metadata.json').exists()



import inspect

# import os
import shutil
from pathlib import Path
from typing import get_type_hints

import pytest

import pharmpy
from pharmpy.deps import numpy as np
from pharmpy.internals.fs.cwd import chdir
from pharmpy.modeling import add_iiv, load_example_model, read_model, set_lower_bounds
from pharmpy.tools.run import (  # retrieve_final_model,; retrieve_models,
    _create_metadata_common,
    _create_metadata_tool,
    _get_run_setup,
    import_tool,
    is_strictness_fulfilled,
    load_example_modelfit_results,
    rank_models,
    read_modelfit_results,
    summarize_errors,
    summarize_modelfit_results,
)
from pharmpy.workflows import LocalDirectoryToolDatabase, local_dask


@pytest.mark.parametrize(
    ('args', 'kwargs'),
    (
        (
            ('ABSORPTION(ZO)', 'exhaustive'),
            {'iiv_strategy': 'no_add'},
        ),
        (
            ('ABSORPTION(ZO)',),
            {'algorithm': 'exhaustive'},
        ),
    ),
)
def test_create_metadata_tool(tmp_path, pheno, args, kwargs):
    with chdir(tmp_path):
        tool_name = 'modelsearch'
        database = LocalDirectoryToolDatabase(tool_name)
        tool = import_tool(tool_name)
        tool_params = inspect.signature(tool.create_workflow).parameters
        tool_param_types = get_type_hints(tool.create_workflow)

        metadata = _create_metadata_tool(
            database=database,
            tool_name=tool_name,
            tool_params=tool_params,
            tool_param_types=tool_param_types,
            args=args,
            kwargs={'model': pheno, **kwargs},
        )

        rundir = tmp_path / 'modelsearch_dir1'

        assert (rundir / 'models' / 'input_model' / '.pharmpy').exists()

        assert metadata['pharmpy_version'] == pharmpy.__version__
        assert metadata['tool_name'] == 'modelsearch'
        assert metadata['tool_options']['model']['__class__'] == 'Model'
        assert metadata['tool_options']['model']['arg_name'] == 'pheno_real'
        assert metadata['tool_options']['model']['db_name'] == 'input_model'
        assert metadata['tool_options']['rank_type'] == 'mbic'
        assert metadata['tool_options']['algorithm'] == 'exhaustive'


def test_create_metadata_tool_raises(tmp_path, pheno):
    with chdir(tmp_path):
        tool_name = 'modelsearch'
        database = LocalDirectoryToolDatabase(tool_name)
        tool = import_tool(tool_name)
        tool_params = inspect.signature(tool.create_workflow).parameters
        tool_param_types = get_type_hints(tool.create_workflow)
        with pytest.raises(Exception, match='modelsearch: \'algorithm\' was not set'):
            _create_metadata_tool(
                database=database,
                tool_name=tool_name,
                tool_params=tool_params,
                tool_param_types=tool_param_types,
                args=('ABSORPTION(ZO)',),
                kwargs={'model': pheno},
            )


def test_get_run_setup(tmp_path):
    with chdir(tmp_path):
        name = 'modelsearch'

        dispatcher, database = _get_run_setup(common_options={}, toolname=name)

        assert dispatcher.__name__.endswith('local_dask')
        assert database.path.stem == 'modelsearch_dir1'

        dispatcher, database = _get_run_setup(
            common_options={'path': 'tool_database_path'}, toolname=name
        )

        assert dispatcher.__name__.endswith('local_dask')
        assert database.path.stem == 'tool_database_path'


def test_create_metadata_common(tmp_path):
    with chdir(tmp_path):
        name = 'modelsearch'

        dispatcher = local_dask
        database = LocalDirectoryToolDatabase(name)

        metadata = _create_metadata_common(
            database=database,
            dispatcher=dispatcher,
            toolname=name,
            common_options={},
        )

        assert metadata['dispatcher'] == 'pharmpy.workflows.dispatchers.local_dask'
        assert metadata['database']['class'] == 'LocalDirectoryToolDatabase'
        path = Path(metadata['database']['path'])
        assert path.stem == 'modelsearch_dir1'
        assert 'path' not in metadata.keys()

        path = 'tool_database_path'

        dispatcher = local_dask
        database = LocalDirectoryToolDatabase(name, path)

        metadata = _create_metadata_common(
            database=database,
            dispatcher=dispatcher,
            toolname=name,
            common_options={'path': path},
        )

        path = Path(metadata['database']['path'])
        assert path.stem == 'tool_database_path'
        assert metadata['path'] == 'tool_database_path'


# def test_retrieve_models(testdata):
#    tool_database_path = testdata / 'results' / 'tool_databases' / 'modelsearch'

#    model_to_retrieve = ['modelsearch_run1']

#    models = retrieve_models(tool_database_path, names=model_to_retrieve)
#    assert len(models) == 1
#    assert models[0].name == model_to_retrieve[0]

#    model_names_all = [
#        'input_model',
#        'modelsearch_run1',
#        'modelsearch_run2',
#        'modelsearch_run3',
#        'modelsearch_run4',
#    ]

#    models = retrieve_models(tool_database_path)
#    assert [model.name for model in models] == model_names_all

#    with open(tool_database_path / 'results.json') as f:
#        results_json = f.read()
#        if os.name == 'nt':
#            new_path = str(tool_database_path).replace('\\', '\\\\')
#        else:
#            new_path = str(tool_database_path)

#        results_json_testpath = results_json.replace('/tmp/tool_results/modelsearch', new_path)

#    res = read_results(results_json_testpath)
#    models = retrieve_models(res, names=model_to_retrieve)
#    assert models[0].name == model_to_retrieve[0]

#    tool_db = LocalDirectoryToolDatabase('modelsearch', path=tool_database_path, exist_ok=True)
#    models = retrieve_models(tool_db, names=model_to_retrieve)
#    assert models[0].name == model_to_retrieve[0]

#    models = retrieve_models(tool_db.model_database, names=model_to_retrieve)
#    assert models[0].name == model_to_retrieve[0]

#    with pytest.raises(ValueError, match='Models {\'x\'} not in database'):
#        retrieve_models(res, names=['x'])

#    tool_without_db_in_results = testdata / 'results' / 'qa_results.json'
#    res = read_results(tool_without_db_in_results)
#    with pytest.raises(
#        ValueError, match='Results type \'QAResults\' does not serialize tool database'
#    ):
#        retrieve_models(res, names=model_to_retrieve)


# def test_retrieve_final_model(testdata):
#    tool_database_path = testdata / 'results' / 'tool_databases' / 'modelsearch'

#    with open(tool_database_path / 'results.json') as f:
#        results_json = f.read()
#        if os.name == 'nt':
#            new_path = str(tool_database_path).replace('\\', '\\\\')
#        else:
#            new_path = str(tool_database_path)

#        results_json_testpath = results_json.replace('/tmp/tool_results/modelsearch', new_path)

#    res = read_results(results_json_testpath)
#    final_model = retrieve_final_model(res)
#    assert final_model.name == 'modelsearch_run2'

#    results_json_none = results_json_testpath.replace(
#        '"final_model_name": "modelsearch_run2"', '"final_model_name": null'
#    )
#    res = read_results(results_json_none)
#    with pytest.raises(ValueError, match='Attribute \'final_model_name\' is None'):
#        retrieve_final_model(res)


def test_summarize_errors(load_model_for_test, testdata, tmp_path, pheno_path):
    with chdir(tmp_path):
        model = read_modelfit_results(pheno_path)
        shutil.copy2(testdata / 'pheno_data.csv', tmp_path)

        error_path = testdata / 'nonmem' / 'errors'

        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path / 'pheno_no_header.mod')
        shutil.copy2(error_path / 'no_header_error.lst', tmp_path / 'pheno_no_header.lst')
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.ext', tmp_path / 'pheno_no_header.ext')
        model_no_header = read_modelfit_results('pheno_no_header.mod')

        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path / 'pheno_rounding_error.mod')
        shutil.copy2(error_path / 'rounding_error.lst', tmp_path / 'pheno_rounding_error.lst')
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.ext', tmp_path / 'pheno_rounding_error.ext')
        model_rounding_error = read_modelfit_results('pheno_rounding_error.mod')

        models = [model, model_no_header, model_rounding_error]
        summary = summarize_errors(models)

        assert 'pheno_real' not in summary.index.get_level_values('model')
        assert len(summary.loc[('pheno_no_header', 'WARNING')]) == 1
        assert len(summary.loc[('pheno_no_header', 'ERROR')]) == 2
        assert len(summary.loc[('pheno_rounding_error', 'ERROR')]) == 2


class DummyModel:
    def __init__(self, name, parameter_names):
        self.name = name
        self.parameters = parameter_names


class DummyResults:
    def __init__(
        self,
        name,
        ofv,
        minimization_successful=True,
        termination_cause=None,
        significant_digits=5,
        warnings=[],
    ):
        self.name = name
        self.ofv = ofv
        self.minimization_successful = minimization_successful
        self.termination_cause = termination_cause
        # 5 is an arbitrary number, this is relevant in test if sig. digits is unreportable (NaN)
        self.significant_digits = significant_digits
        self.warnings = warnings


def test_rank_models():
    base = DummyModel('base', parameter_names=['p1'])
    m1 = DummyModel('m1', parameter_names=['p1', 'p2'])
    m2 = DummyModel('m2', parameter_names=['p1', 'p2'])
    m3 = DummyModel('m3', parameter_names=['p1', 'p2', 'p3'])
    m4 = DummyModel('m4', parameter_names=['p1'])

    base_res = DummyResults(name=base.name, ofv=0)
    m1_res = DummyResults(
        name=m1.name, ofv=-5, minimization_successful=False, termination_cause='rounding_errors'
    )
    m2_res = DummyResults(name=m2.name, ofv=-4)
    m3_res = DummyResults(name=m3.name, ofv=-4)
    m4_res = DummyResults(name=m4.name, ofv=1)

    models = [m1, m2, m3, m4]
    models_res = [m1_res, m2_res, m3_res, m4_res]

    df = rank_models(base, base_res, models, models_res, rank_type='ofv')
    assert len(df) == 5
    best_model = df.loc[df['rank'] == 1].index.values
    assert list(best_model) == ['m2', 'm3']

    # Test if rounding errors are allowed
    df = rank_models(
        base,
        base_res,
        models,
        models_res,
        strictness='minimization_successful or (rounding_errors and sigdigs>=0)',
        rank_type='ofv',
    )
    best_model = df.loc[df['rank'] == 1].index.values
    assert list(best_model) == ['m1']
    ranked_models = df.dropna().index.values
    assert len(ranked_models) == 5

    # Test with a cutoff of dOFV=1
    df = rank_models(base, base_res, models, models_res, rank_type='ofv', cutoff=1)
    ranked_models = df.dropna().index.values
    assert len(ranked_models) == 3

    # Test with LRT
    df = rank_models(base, base_res, models, models_res, rank_type='lrt', cutoff=0.05)
    ranked_models = list(df.dropna().index.values)
    assert sorted(ranked_models) == ['base', 'm2']

    df = rank_models(
        base,
        base_res,
        models,
        models_res,
        parent_dict={m: base for m in models},
        rank_type='lrt',
        cutoff=0.05,
    )
    ranked_models = list(df.dropna().index.values)
    assert sorted(ranked_models) == ['base', 'm2']

    # Test if candidate model does not have an OFV
    m5 = DummyModel('m5', parameter_names=['p1'])
    m5_res = DummyResults(name=m5.name, ofv=np.nan)
    df = rank_models(base, base_res, models + [m5], models_res + [m5_res], rank_type='ofv')
    ranked_models = list(df.dropna().index.values)
    assert 'm5' not in ranked_models
    assert np.isnan(df.loc['m5']['rank'])

    # Test if model has minimized but has unreportable number of significant digits while still allowing rounding
    # errors
    m6 = DummyModel('m6', parameter_names=['p1'])
    m6_res = DummyResults(
        name=m6.name,
        ofv=-5,
        minimization_successful=False,
        termination_cause='rounding_errors',
        significant_digits=np.nan,
    )
    df = rank_models(
        base,
        base_res,
        models + [m6],
        models_res + [m6_res],
        strictness='minimization_successful or (rounding_errors and sigdigs>=0)',
        rank_type='ofv',
    )
    ranked_models = list(df.dropna().index.values)
    assert 'm6' not in ranked_models
    assert np.isnan(df.loc['m6']['rank'])

    # Test if base model failed, fall back to rank value
    base_nan = DummyModel('base_nan', parameter_names=['p1'])
    base_nan_res = DummyResults(name=base_nan.name, ofv=np.nan)
    df = rank_models(
        base_nan,
        base_nan_res,
        models,
        models_res,
        strictness='minimization_successful or (rounding_errors and sigdigs>=0)',
        rank_type='ofv',
    )
    assert df.iloc[0].name == 'm1'

    # Test if base model failed but still has OFV with a very high value, fall back to rank value
    base_nan = DummyModel(
        'base_nan',
        parameter_names=['p1'],
    )
    base_nan_res = DummyResults(
        name=base_nan.name, ofv=2e154, minimization_successful=False, termination_cause='something'
    )
    df = rank_models(
        base_nan,
        base_nan_res,
        models,
        models_res,
        strictness='minimization_successful or (rounding_errors and sigdigs>=0)',
        rank_type='ofv',
    )
    assert df.iloc[0].name == 'm1'
    assert df.nunique()['ofv'] == df.nunique()['rank']

    with pytest.raises(ValueError):
        rank_models(base, base_res, models + [m5], models_res)


def test_rank_models_bic(load_model_for_test, testdata):
    model_base = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model_iiv = add_iiv(model_base, ['S1'], 'exp')
    model_iiv = model_iiv.replace(name='pheno_iiv')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno.mod')
    df_mixed = rank_models(model_base, res, [model_iiv], [res], rank_type='bic', bic_type='mixed')
    df_mult_test = rank_models(
        model_base,
        res,
        [model_iiv],
        [res],
        rank_type='bic',
        bic_type='mixed',
        multiple_testing=True,
        mult_test_p=2,
    )
    assert df_mixed.loc['pheno_iiv', 'bic'] != df_mult_test.loc['pheno_iiv', 'bic']


def test_summarize_modelfit_results(
    load_model_for_test, create_model_for_test, testdata, pheno_path
):
    pheno = read_modelfit_results(pheno_path)

    summary_single = summarize_modelfit_results(pheno)

    assert summary_single.loc['pheno_real']['ofv'] == 586.2760562818805
    assert summary_single['IVCL_estimate'].mean() == 0.0293508

    assert len(summary_single.index) == 1

    mox = read_modelfit_results(testdata / 'nonmem' / 'models' / 'mox1.mod')

    summary_multiple = summarize_modelfit_results([pheno, mox])

    assert summary_multiple.loc['mox1']['ofv'] == -624.5229577248352
    assert summary_multiple['IIV_CL_estimate'].mean() == 0.41791
    assert summary_multiple['IIV_CL_V_estimate'].mean() == 0.395647  # One is NaN

    assert len(summary_multiple.index) == 2
    assert list(summary_multiple.index) == ['pheno_real', 'mox1']

    summary_no_res = summarize_modelfit_results([pheno, None])

    assert summary_no_res.loc['pheno_real']['ofv'] == 586.2760562818805

    pheno_multest = read_modelfit_results(
        testdata
        / 'nonmem'
        / 'modelfit_results'
        / 'onePROB'
        / 'multEST'
        / 'noSIM'
        / 'pheno_multEST.mod'
    )

    summary_multest = summarize_modelfit_results([pheno_multest, mox])

    assert len(summary_multest.index) == 2

    assert not summary_multest.loc['pheno_multEST']['minimization_successful']
    summary_multest_full = summarize_modelfit_results(
        [pheno_multest, mox], include_all_estimation_steps=True
    )

    assert len(summary_multest_full.index) == 3
    assert len(set(summary_multest_full.index.get_level_values('model'))) == 2
    assert summary_multest_full.loc['pheno_multEST', 1]['run_type'] == 'estimation'
    assert summary_multest_full.loc['pheno_multEST', 2]['run_type'] == 'evaluation'

    assert not summary_multest_full.loc['pheno_multEST', 1]['minimization_successful']

    summary_multest_full_no_res = summarize_modelfit_results(
        [None, mox],
        include_all_estimation_steps=True,
    )

    assert summary_multest_full_no_res.loc['mox1', 1]['ofv'] == -624.5229577248352

    with pytest.raises(ValueError, match='Option `results` is None'):
        summarize_modelfit_results(None)

    with pytest.raises(ValueError, match='All input results are empty'):
        summarize_modelfit_results([None, None])


def test_summarize_modelfit_results_errors(load_model_for_test, testdata, tmp_path, pheno_path):
    with chdir(tmp_path):
        model = read_modelfit_results(pheno_path)
        shutil.copy2(testdata / 'pheno_data.csv', tmp_path)

        error_path = testdata / 'nonmem' / 'errors'

        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path / 'pheno_no_header.mod')
        shutil.copy2(error_path / 'no_header_error.lst', tmp_path / 'pheno_no_header.lst')
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.ext', tmp_path / 'pheno_no_header.ext')
        model_no_header = read_modelfit_results('pheno_no_header.mod')

        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path / 'pheno_rounding_error.mod')
        shutil.copy2(error_path / 'rounding_error.lst', tmp_path / 'pheno_rounding_error.lst')
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.ext', tmp_path / 'pheno_rounding_error.ext')
        model_rounding_error = read_modelfit_results('pheno_rounding_error.mod')

        results = [
            model,
            model_no_header,
            model_rounding_error,
        ]
        summary = summarize_modelfit_results(results)

        assert summary.loc['pheno_real']['errors_found'] == 0
        assert summary.loc['pheno_real']['warnings_found'] == 0
        assert summary.loc['pheno_no_header']['errors_found'] == 2
        assert summary.loc['pheno_no_header']['warnings_found'] == 1
        assert summary.loc['pheno_rounding_error']['errors_found'] == 2
        assert summary.loc['pheno_rounding_error']['warnings_found'] == 0


def test_read_modelfit_results(testdata):
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    assert res.ofv == 586.27605628188053
    expected_rse = {
        'PTVCL': 0.04473086219931638,
        'PTVV': 0.027325355750219965,
        'THETA_3': 0.5270721117543418,
        'IVCL': 0.4570676097414721,
        'IVV': 0.2679176521178241,
        'SIGMA_1_1': 0.17214711879767391,
    }
    assert res.relative_standard_errors.to_dict() == expected_rse

    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_design.mod')
    assert res.ofv == 730.9699060285753
    expected_rse = {
        'TVCL': 0.07540090004436839,
        'TVV': 0.06030486531634996,
        'IVCL': 0.3895984361000244,
        'IVV': 0.19632785003247685,
        'SIGMA_1_1': 0.1865063481541913,
    }
    assert res.relative_standard_errors.to_dict() == expected_rse


def test_load_example_modelfit_results():
    res = load_example_modelfit_results("pheno")
    assert res.ofv == 586.27605628188053


@pytest.mark.parametrize(
    'path, statement, expected',
    [
        (
            'nonmem/modelfit_results/onePROB/oneEST/noSIM/near_bounds.mod',
            'minimization_successful or (rounding_errors and sigdigs > 0)',
            True,
        ),
        (
            'nonmem/pheno_real.mod',
            'condition_number < 1000',
            False,
        ),
        (
            'nonmem/pheno_real.mod',
            'condition_number < 300000',
            True,
        ),
        (
            'nonmem/modelfit_results/onePROB/oneEST/noSIM/maxeval3.mod',
            'rse < 30',
            True,
        ),
        (
            'nonmem/pheno.mod',
            'minimization_successful and sigdigs > 0 and rse<2',
            False,
        ),
        (
            'nonmem/pheno.mod',
            'minimization_successful and sigdigs > 3',
            True,
        ),
        (
            'nonmem/pheno.mod',
            'final_zero_gradient',
            False,
        ),
        (
            'nonmem/pheno.mod',
            'final_zero_gradient_theta',
            False,
        ),
        (
            'nonmem/pheno.mod',
            'final_zero_gradient_omega',
            False,
        ),
        (
            'nonmem/pheno.mod',
            'final_zero_gradient_sigma',
            False,
        ),
    ],
)
def test_strictness(testdata, path, statement, expected):
    res = read_modelfit_results(testdata / path)
    model = read_model(testdata / path)
    assert is_strictness_fulfilled(res, model, statement) == expected


def test_strictness_unallowed_operators(testdata):
    res = read_modelfit_results(testdata / 'nonmem/pheno.mod')
    model = read_model(testdata / 'nonmem/pheno.mod')
    with pytest.raises(ValueError, match=r"Unallowed operators found: &"):
        is_strictness_fulfilled(res, model, 'minimization_successful & rounding_errors')
    with pytest.raises(ValueError, match=r"Unallowed operators found: &, |"):
        is_strictness_fulfilled(
            res, model, 'minimization_successful & (rounding_errors | sigdigs>3)'
        )


def test_strictness_parameters(testdata):
    res = load_example_modelfit_results('pheno')
    model = load_example_model("pheno")
    assert not is_strictness_fulfilled(res, model, 'rse_theta < 0.3')
    assert is_strictness_fulfilled(res, model, 'rse_theta < 0.55')
    assert not is_strictness_fulfilled(res, model, 'rse_omega < 0.3')
    assert is_strictness_fulfilled(res, model, 'rse_omega < 0.5')
    assert is_strictness_fulfilled(res, model, 'rse_sigma < 0.2')

    res = read_modelfit_results(testdata / 'nonmem/pheno.mod')
    model = read_model(testdata / 'nonmem/pheno.mod')
    assert not is_strictness_fulfilled(res, model, 'estimate_near_boundary_theta')
    model = set_lower_bounds(model, {'TVCL': 0.0058})
    assert is_strictness_fulfilled(res, model, 'estimate_near_boundary_theta')



from dataclasses import dataclass
from typing import Any, Optional

from pharmpy.model import Model
from pharmpy.tools.external.nonmem.results import simfit_results
from pharmpy.workflows import Results


@dataclass(frozen=True)
class SimfitResults(Results):
    """Simfit results class"""

    modelfit_results: Optional[Any] = None


def calculate_results(modelfit_results):
    """Calculate simfit results"""
    return SimfitResults(modelfit_results=modelfit_results)


def psn_simfit_results(paths):
    modelfit_results = []
    for path in paths:
        model = Model.parse_model(path)
        modelfit_results.extend(simfit_results(model, path))
    res = calculate_results(modelfit_results)
    return res



from importlib.metadata import version
from typing import Iterable, Iterator, List, Tuple, Union

from lark import Token, Transformer, Tree
from lark.tree import Meta

WS = {' ', '\x00', '\t'}
LF = {'\r', '\n'}


def _tokenize_ignored_characters(s: str, i: int, j: int) -> Iterable[Token]:
    # TODO: Propagate line/column information
    head = i
    while i < j:
        first = s[head]
        head += 1
        if first in WS:
            while head < j and s[head] in WS:
                head += 1

            yield Token('WS', s[i:head], start_pos=i, end_pos=head)

        elif first == ';':
            while head < j and s[head] not in LF:
                head += 1

            yield Token('COMMENT', s[i:head], start_pos=i, end_pos=head)

        elif first == '\r':
            assert head < j and s[head] == '\n'
            head += 1

            yield Token('NEWLINE', s[i:head], start_pos=i, end_pos=head)

        elif first == '\n':
            yield Token('NEWLINE', s[i:head], start_pos=i, end_pos=head)

        else:
            assert first == '&'
            while head < j and s[head] not in LF:
                head += 1

            yield Token('CONT', s[i:head], start_pos=i, end_pos=head)

        i = head


def _item_range(x: Union[Tree, Token]) -> Tuple[int, int]:
    if isinstance(x, Tree):
        i = x.meta.start_pos
        j = x.meta.end_pos
        if version('lark') == '1.1.6':
            j = _get_new_end_pos(x)
    else:
        i = x.start_pos
        j = x.end_pos

    assert isinstance(i, int)
    assert isinstance(j, int)
    return (i, j)


def _get_new_end_pos(x) -> int:
    # FIXME: Temporary workaround, see https://github.com/lark-parser/lark/issues/1304
    x = x.children[-1]
    if isinstance(x, Token):
        endpos = x.end_pos
        assert isinstance(endpos, int)
        return endpos
    return _get_new_end_pos(x)


def _interleave_ignored(source: str, it: Iterator[Union[Tree, Token]]):
    x = next(it)
    yield x

    i = _item_range(x)[1]

    x = next(it)
    while True:
        j, k = _item_range(x)

        if i < j:
            yield from _tokenize_ignored_characters(source, i, j)

        yield x

        try:
            x = next(it)
        except StopIteration:
            break

        i = k


def interleave_ignored(source: str, children: List[Union[Tree, Token]]):
    return children if len(children) < 2 else list(_interleave_ignored(source, iter(children)))


class InterleaveIgnored(Transformer):
    def __init__(self, source: str):
        self.source = source

    def __default__(self, data, children, meta):
        new_children = interleave_ignored(self.source, children)
        return Tree(data, new_children, meta)


def with_ignored_tokens(source, tree):
    new_tree = InterleaveIgnored(source).transform(tree)

    if version('lark') == '1.1.6' and new_tree.children:
        new_tree.meta.end_pos = _get_new_end_pos(new_tree)

    final_meta = Meta()
    # TODO: Propagate line/column information
    final_meta.start_pos = 0
    final_meta.end_pos = len(source)
    final_meta.empty = False

    final_children: List[Union[Tree, Token]] = (
        (
            list(_tokenize_ignored_characters(source, 0, new_tree.meta.start_pos))
            + new_tree.children
            + list(_tokenize_ignored_characters(source, new_tree.meta.end_pos, len(source)))
        )
        if new_tree.children
        else list(_tokenize_ignored_characters(source, 0, len(source)))
    )

    final_data = new_tree.data
    return Tree(final_data, final_children, final_meta)



from itertools import chain, combinations
from typing import Iterable, Tuple, TypeVar

T = TypeVar('T')


def subsets(iterable: Iterable[T], min_size: int = 0, max_size: int = -1) -> Iterable[Tuple[T]]:
    """Returns an iterable over all the subsets of the input iterable with
    minimum and maximum size constraints. Allows maximum_size to be given
    relatively to iterable "length" by specifying a negative value.

    Adapted from powerset function defined in
    https://docs.python.org/3/library/itertools.html#itertools-recipes

    subsets([1,2,3], min_size=1, max_size=2) --> (1,) (2,) (3,) (1,2) (1,3) (2,3)"
    """
    s = list(iterable)
    max_size = len(s) + max_size + 1 if max_size < 0 else max_size
    combs = (combinations(s, r) for r in range(min_size, max_size + 1))
    return chain.from_iterable(combs)  # pyright: ignore [reportGeneralTypeIssues]


def non_empty_proper_subsets(iterable: Iterable[T]) -> Iterable[Tuple[T]]:
    """Returns an iterable over all the non-empty proper subsets of the input
    iterable.

    non_empty_proper_subsets([1,2,3]) --> (1,) (2,) (3,) (1,2) (1,3) (2,3)"
    """
    return subsets(iterable, min_size=1, max_size=-2)


def non_empty_subsets(iterable: Iterable[T]) -> Iterable[Tuple[T]]:
    """Returns an iterable over all the non-empty subsets of the input
    iterable.

    non_empty_subsets([1,2,3]) --> (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    """
    return subsets(iterable, min_size=1, max_size=-1)



"""
The NONMEM $SUBROUTINES record
"""

from .option_record import OptionRecord


class SubroutineRecord(OptionRecord):
    @property
    def advan(self):
        advan = self.get_option('ADVAN')
        if advan is None:
            advan = self.get_option_startswith('ADVAN')
        return advan

    def set_advan(self, value):
        newrec = self.remove_option_startswith('ADVAN').append_option(value)
        return newrec

    @property
    def trans(self):
        trans = self.get_option('TRANS')
        if trans is None:
            trans = self.get_option_startswith('TRANS')
        if trans is None:
            trans = 'TRANS1'
        return trans

    @property
    def tol(self):
        tol = self.get_option('TOL')
        if tol is not None:
            tol = int(tol)
        return tol

    @property
    def atol(self):
        atol = self.get_option('ATOL')
        if atol is None:
            atol = 1.0e-12
        else:
            atol = float(atol)
        return atol



import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Literal, Optional, Union

import pharmpy.tools.estmethod.algorithms as algorithms
from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.internals.fn.signature import with_same_arguments_as
from pharmpy.internals.fn.type import with_runtime_arguments_type_check
from pharmpy.model import Model
from pharmpy.modeling import has_linear_odes
from pharmpy.tools import summarize_errors, summarize_modelfit_results
from pharmpy.tools.common import ToolResults
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder
from pharmpy.workflows.results import ModelfitResults

METHODS = ('FOCE', 'FO', 'IMP', 'IMPMAP', 'ITS', 'SAEM', 'LAPLACE', 'BAYES')
SOLVERS = ('CVODES', 'DGEAR', 'DVERK', 'IDA', 'LSODA', 'LSODI')
PARAMETER_UNCERTAINTY_METHODS = ('SANDWICH', 'CPG', 'OFIM')

ALGORITHMS = frozenset(['exhaustive', 'exhaustive_with_update', 'exhaustive_only_eval'])


def create_workflow(
    algorithm: Literal[tuple(ALGORITHMS)],
    methods: Optional[Union[List[Literal[METHODS]], Literal['all']]] = None,
    solvers: Optional[Union[List[Literal[SOLVERS]], Literal[SOLVERS]]] = None,
    parameter_uncertainty_methods: Optional[
        Union[List[Literal[PARAMETER_UNCERTAINTY_METHODS]], Literal[PARAMETER_UNCERTAINTY_METHODS]]
    ] = None,
    results: Optional[ModelfitResults] = None,
    model: Optional[Model] = None,
):
    """Run estmethod tool.

    Parameters
    ----------
    algorithm : str
         The algorithm to use (can be 'exhaustive', 'exhaustive_with_update' or 'exhaustive_only_eval')
    methods : list of {'FOCE', 'FO', 'IMP', 'IMPMAP', 'ITS', 'SAEM', 'LAPLACE', 'BAYES'}, None or 'all'
         List of estimation methods to test.
         Can be specified as 'all', a list of estimation methods, or None (to not test any estimation method)
    solvers : str or list of {'CVODES', 'DGEAR', 'DVERK', 'IDA', 'LSODA', 'LSODI'} or None
         List of solver to test. Can be specified as 'all', a list of solvers, or None (to
         not test any solver)
    parameter_uncertainty_methods : str or list of {'SANDWICH', 'CPG', 'OFIM'} or None
         List of parameter uncertainty methods to test.
         Can be specified as 'all', a list of uncertainty methods, or None (to not evaluate any uncertainty)
    results : ModelfitResults
         Results for model
    model : Model
         Pharmpy model

    Returns
    -------
    EstMethodResults
        Estmethod tool result object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> from pharmpy.tools import run_estmethod, load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> methods = ['IMP', 'SAEM']
    >>> parameter_uncertainty_methods = None
    >>> run_estmethod( # doctest: +SKIP
    >>>     'reduced', methods=methods, solvers='all', # doctest: +SKIP
    >>>      parameter_uncertainty_methods=parameter_uncertainty_methods, results=results, model=model # doctest: +SKIP
    >>> ) # doctest: +SKIP

    """
    wb = WorkflowBuilder(name="estmethod")

    algorithm_func = getattr(algorithms, algorithm)

    if model is not None:
        start_task = Task('start_estmethod', start, model, results)
    else:
        start_task = Task('start_estmethod', start)

    wb.add_task(start_task)

    if methods is None:
        methods = [model.estimation_steps[-1].method]

    wf_algorithm, task_base_model_fit = algorithm_func(
        _format_input(methods, METHODS),
        _format_input(solvers, SOLVERS),
        _format_input(parameter_uncertainty_methods, PARAMETER_UNCERTAINTY_METHODS),
    )
    wb.insert_workflow(wf_algorithm, predecessors=start_task)

    wf_fit = create_fit_workflow(n=len(wb.output_tasks))
    wb.insert_workflow(wf_fit, predecessors=wb.output_tasks)

    model_tasks = [start_task] + wb.output_tasks
    if task_base_model_fit:
        model_tasks.extend(task_base_model_fit)

    task_post_process = Task('post_process', post_process)
    wb.add_task(task_post_process, predecessors=model_tasks)

    return Workflow(wb)


def _format_input(input_option, default_option):
    if input_option == 'all':
        return default_option
    elif input_option is None:
        return [None]
    else:
        return [entry.upper() for entry in input_option]


def start(model, modelfit_results):
    log = modelfit_results.log if modelfit_results else None
    model_entry = ModelEntry(model, modelfit_results=modelfit_results, log=log)
    return model_entry


def post_process(*model_entries):
    summary_tool = summarize_tool(model_entries)

    res_models = [model_entry.model for model_entry in model_entries]
    res_modelfit_results = [model_entry.modelfit_results for model_entry in model_entries]

    summary_models = summarize_modelfit_results(
        res_modelfit_results,
        include_all_estimation_steps=True,
    )
    summary_errors = summarize_errors(res_modelfit_results)
    summary_settings = summarize_estimation_steps(res_models)

    return EstMethodResults(
        summary_tool=summary_tool,
        summary_models=summary_models,
        summary_settings=summary_settings,
        summary_errors=summary_errors,
    )


@dataclass(frozen=True)
class EstMethodResults(ToolResults):
    rst_path = Path(__file__).resolve().parent / 'report.rst'

    summary_settings: Optional[Any] = None


def summarize_tool(model_entries):
    rows = {}

    for model_entry in model_entries:
        model, res = model_entry.model, model_entry.modelfit_results
        description, parent_model = model.description, model.parent_model
        if res is not None:
            ofv = res.ofv
            runtime_est = res.estimation_runtime_iterations.iloc[0]
        else:
            ofv, runtime_est = np.nan, np.nan
        rows[model.name] = (description, ofv, runtime_est, parent_model)

    colnames = ['description', 'ofv', 'runtime_estimation', 'parent_model']
    index = pd.Index(rows.keys(), name='model')
    df = pd.DataFrame(rows.values(), index=index, columns=colnames)

    return df.sort_values(by=['ofv'])


def summarize_estimation_steps(models):
    dfs = {}
    for model in models:
        df = model.estimation_steps.to_dataframe()
        df.index = range(1, len(df) + 1)
        dfs[model.name] = df.drop(columns=['tool_options'])

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="The behavior of DataFrame concatenation with empty or all-NA",
            category=FutureWarning,
        )
        summary = pd.concat(list(dfs.values()), keys=list(dfs.keys()))
    return summary


@with_runtime_arguments_type_check
@with_same_arguments_as(create_workflow)
def validate_input(algorithm, methods, solvers, parameter_uncertainty_methods, model):
    if solvers is not None and has_linear_odes(model):
        raise ValueError(
            'Invalid input `model`: testing non-linear solvers on linear system is not supported'
        )

    if methods is None and solvers is None and parameter_uncertainty_methods is None:
        raise ValueError(
            'Invalid search space options: please specify at least '
            'one of `methods`, `solvers`, or `parameter_uncertainty_methods`'
        )

    if methods is not None:
        _validate_search_space(methods, METHODS, 'methods')

    if solvers is not None:
        _validate_search_space(solvers, SOLVERS, 'solvers')

    if parameter_uncertainty_methods is not None:
        _validate_search_space(
            parameter_uncertainty_methods,
            PARAMETER_UNCERTAINTY_METHODS,
            'parameter_uncertainty_methods',
        )


def _validate_search_space(input_search_space, allowed_search_space, option_name):
    if isinstance(input_search_space, str):
        if input_search_space != 'all':
            raise ValueError(
                f'Invalid `{option_name}`: if option is str it must be `all`, got {input_search_space}'
            )
    else:
        option_diff = {option.upper() for option in input_search_space}.difference(
            allowed_search_space
        )
        if option_diff:
            raise ValueError(
                f'Invalid `{option_name}`: {option_diff} not in {sorted(allowed_search_space)}'
            )
